from .context import Context

from typing import *

#exceptions
class Exceptions:
	#customizable
	class InvalidInteger(Exception):
		pass

	class InvalidBoolean(Exception):
		pass

	class InvalidRole(Exception):
		pass

	class InvalidMention(Exception):
		pass

	class InvalidChannel(Exception):
		pass

	class MissingArguments(Exception):
		pass

	#not customizable
	class FileError(Exception):
		pass

	class InvalidPermission(Exception):
		pass

	#http exceptions
	class BadRequest(Exception): #400
		pass

	class NotAllowed(Exception): #401
		pass

	class InvalidPermissions(Exception): #403
		pass

	class NotFound(Exception): #404
		pass

	class InternalServerError(Exception): #>499
		pass

	class UnknownError(Exception): #i forgot a code or something
		pass

#inherit, customize, pass through client init
class CustomizableExceptions:
	#type conversion exceptions
	async def onInvalidInteger(self,arg:str,context:Context) -> None:
		raise Exceptions.InvalidInteger(f"Provided argument '{arg}' is not an integer.")

	async def onInvalidBoolean(self,arg:str,context:Context) -> None:
		raise Exceptions.InvalidBoolean(f"Provided argument '{arg}' is not a boolean.")

	async def onInvalidRole(self,arg:str,context:Context) -> None:
		raise Exceptions.InvalidRole(f"Provided argument '{arg}' is not a role mention.\033[0m")

	async def onInvalidMention(self,arg:str,context:Context) -> None:
		raise Exceptions.InvalidMention(f"Provided argument '{arg}' is not a mention.\033[0m")

	async def onInvalidChannel(self,arg:str,context:Context) -> None:
		raise Exceptions.InvalidChannel(f"Provided argument '{arg}' is not a channel.\033[0m")

	#when user dont know how to use command
	async def onMissingArguments(self,args:List[str],context:Context) -> None:
		raise Exceptions.MissingArguments(f"Argument(s) \"{', '.join(args)}\" are missing from message.\033[0")



from typing import *

class CustomStatus:
	def __init__(self,status:str="online",type:int=1,game:Union[str,None]=None,url:Union[str,None]=None):
		self.status = status
		self.type = type
		self.game = game
		self.url = url

	def _toJson(self):
		data ={'afk':False,'since':None,'status': self.status,'game': {"name":self.game,"type":self.type}}
		if self.url:
			data['game']['url'] = self.url
		return data

class ClientSettings:
	def __init__(self,logging:bool=True,reply_to_self:bool=False,reply_to_bots:bool=False,apply_to_events:bool=True,split_by_quotations:bool=True):
		self.split_by_quotations = split_by_quotations
		self.apply_to_events = apply_to_events
		self.reply_to_bots = reply_to_bots
		self.reply_to_self = reply_to_self
		self.logging = logging


from PIL.Image import open as openImage
from io import BytesIO
import fake_useragent
from typing import *
import requests

def getDimensions(image_url):
	response = requests.get(image_url,headers={"User-Agent":fake_useragent.UserAgent(browsers=['chrome', 'firefox', 'opera', 'safari', 'edge', 'internet explorer']).random})
	if response.status_code == 200:
		img = openImage(BytesIO(response.content))
		width, height = img.size
		return width, height, True
	else:
		raise Exception(f"Failed to retrieve image from URL. Status code: {response.status_code}")

class Field:
	def __init__(self,name,value,inline=False):
		self.name = name
		self.value = value
		self.inline = inline

class Footer:
	def __init__(self,text,icon_url=None):
		self.text = text
		self.icon_url = icon_url

class EmbedAuthor:
	def __init__(self,name,url=None,icon_url=None):
		self.name = name
		self.url = url
		self.icon_url = icon_url

class Thumbnail:
	def __init__(self,url,width=None,height=None):
		self.url = url
		self.width = width
		self.height = height
		self.available = False

	async def getDimensions(self):
		self.width,self.height,self.available = getDimensions(self.url)

class Image:
	def __init__(self,url):
		self.url = url
		self.width = None
		self.height = None
		self.available = False

	async def getDimensions(self):
		self.width,self.height,self.available = getDimensions(self.url)

class Provider:
	def __init__(self,data):
		self.name = data['name']
		self.url = data.get('url')

class Video:
	def __init__(self,url,width=None,height=None):
		self.url = url
		self.width = width
		self.height = height
		self.available = False

	async def getDimensions(self):
		self.width,self.height,self.available = getDimensions(self.url)

class Embed:
	def __init__(self,title:str,description:str,color:int=0,url:Union[str,None]=None,timestamp:Union[str,None]=None):
		self.title = title
		self.description = description
		self.color = color
		self.fields = []
		self.url = url
		self.timestamp = timestamp
		self.author = None
		self.footer = None
		self.thumbnail = None
		self.image = None

	async def addAuthor(self,name:str,url:Union[str,None]=None,icon_url:Union[str,None]=None):
		self.author = EmbedAuthor(name,url,icon_url)

	async def addFooter(self,text:str,icon_url:Union[str,None]=None):
		self.footer = Footer(text,icon_url)

	async def addField(self,name:str,value:str,inline:bool=False):
		self.fields.append(Field(name,value,inline))

	async def addThumbnail(self,url:str):
		self.thumbnail = Thumbnail(url)

	async def addImage(self,url:str):
		self.image = Image(url)

	#for usage in the internal folder
	def _addAuthor(self,name,url=None,icon_url=None):
		self.author = EmbedAuthor(name,url,icon_url)
	
	def _addFooter(self,text,icon_url=None):
		self.footer = Footer(text,icon_url)
	
	def _addField(self,name,value,inline=False):
		self.fields.append(Field(name,value,inline))
	
	def _addThumbnail(self,url):
		self.thumbnail = Thumbnail(url)
	
	def _addImage(self,url):
		self.image = Image(url)

	def _addVideo(self,url,width=None,height=None):
		self.video = Video(url,width,height)

	def _addProvider(self,data):
		self.provider = Provider(data)

class ImageEmbed:
	def __init__(self,data):
		self.type = data['type']
		self.url = data['url']
		self.thumbnail = Thumbnail(data['thumbnail']['url'],data['thumbnail']['width'],data['thumbnail']['height'])

	def _addThumbnail(self,url):
		self.thumbnail = Thumbnail(url)
	
	def _addImage(self,url):
		self.image = Image(url)
	
	def _addVideo(self,url,width=None,height=None):
		self.video = Video(url,width,height)
	
	def _addProvider(self,data):
		self.provider = Provider(data)


from setuptools import find_packages, setup

"""
python setup.py sdist bdist_wheel
twine upload dist/*
"""

setup(
    name='hmus',
    version='1.1.3',
    author='LG125YT',
    author_email='lg125yt@gmail.com',
    description='An asynchronous API wrapper for Hummus by Ziad87',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/LG125YT/Hummus.py',
    packages=find_packages(),
    install_requires=[
        "datetime",
        "asyncio",
        "fake_useragent",
        "websockets",
        "requests",
        "requests_toolbelt",
        "filetype",
        "Pillow",
        "typing"
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
    ],
)



from .file import File, Attachment
from .embed import Embed

from typing import *
import re

class Message:
	def __init__(self,data,instance,reply:bool=False,reply_content=None,reply_author=None):
		from .internal import makeEmbed
		from .guild import Guild
		from .role import Role
		from .user import User
		if reply:
			self.is_reply:bool = True
			self.original_reply = reply_content
			self.original_author = reply_author
		else:
			self.is_reply:bool = False
		self.instance = instance
		self.guild_id:Union[str,None] = data.get("guild_id")
		self.author:User = User(data['author'],instance,self.guild_id)
		self.content:str = data['content']
		self.channel_id:str = data['channel_id']
		self.mention_roles:list[Role] = []
		self.guild:Union[Guild,None] = None
		self.channel:Channel = None #type:ignore
		if self.guild_id:
			for guild in instance.guilds:
				if self.guild_id == guild.id:
					self.guild = guild
					for channel in guild.channels:
						if data['channel_id'] == channel.id:
							self.channel = channel
					for r in guild.roles:
						if r.id in data['mention_roles']:
							self.mention_roles.append(r)
		else:
			for dm in self.instance.state.private_channels:
				if data['channel_id'] == dm.id:
					self.channel = dm
		self.dm:bool = False
		if not self.guild_id:
			self.dm = True
		self.id:str = data['id']
		self.mentions:list[User] = [User(mention,instance,self.guild_id) for mention in data['mentions']]
		self.attachments:list[Attachment] = [Attachment(attachment,self.instance) for attachment in data['attachments']]
		self.embeds:list[Embed] = makeEmbed(data['embeds'])
		self.edited_timestamp:Union[str,None] = data['edited_timestamp']
		self.pinned:bool = data['pinned']
		self.mention_everyone:bool = data['mention_everyone']
		self.tts:bool = data['tts']
		self.timestamp:str = data['timestamp']
		self.webhook_id:Union[str,None] = data['webhook_id']

	async def typing(self) -> None:
		return await self.instance.http.start_typing(self.channel_id)

	async def delete(self) -> None:
		return await self.instance.http.delete_message(self.channel_id,self.id)

	async def edit(self,content:str) -> 'Message':
        filtered = re.sub(r'<@!?(\d+)>|@everyone|<@&(\d+)>', '[mention]', self.original_reply)
		if self.is_reply:
			content = f"> {filtered}\n<@{self.original_author.id}> {content}"
		return await self.instance.http.edit_message(self.channel_id,self.id,content,_reply=self.is_reply,_reply_content=self.original_reply,_reply_author=self.original_author)

	async def send(self,content:str="", embeds:List[Embed]=[],file:Union[File,None]=None) -> 'Message':
		return await self.instance.http.send_message(self.channel_id,content,embeds,file)

	async def reply(self,content:str="",embeds:List[Embed]=[],file:Union[File,None]=None) -> 'Message':
		from .internal import Reply
		filtered = re.sub(r'<@!?(\d+)>|@everyone|<@&(\d+)>', '[mention]', self.content)
		return await self.instance.http.send_message(self.channel_id,f"> {filtered}\n<@{self.author.id}> {content}",embeds,file,_reply=Reply(self))

	async def pin(self) -> None:
		return await self.instance.http.pin_message(self.channel_id,self.id)

	async def unpin(self) -> None:
		return await self.instance.http.unpin_message(self.channel_id,self.id)



from ...message import Message
from ...embed import Embed
from ...user import User
from ...file import File

from ..message import prepareEmbed, Reply

from requests_toolbelt import MultipartEncoder
from typing import *

import random
import string
import json

class hMessage:
	def __init__(self,instance):
		self.instance = instance
		self.s = instance.s

	async def start_typing(self,channel_id:str) -> None:
		from ... import HTTPStatus
		r = self.s.post(f"{self.instance.base_url}/channels/{channel_id}/typing")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def delete_message(self,channel_id:str,message_id:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(url=f"{self.instance.base_url}channels/{channel_id}/messages/{message_id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def edit_message(self,channel_id:str, id:str,content:str,_reply:bool=False,_reply_content:Union[str,None]=None,_reply_author:Union[User,None]=None) -> Message:
		from ... import HTTPStatus
		r = self.s.patch(f"{self.instance.base_url}channels/{channel_id}/messages/{id}",json={"content":content})
		s = HTTPStatus(r)
		if s.success:
			return Message(r.json(),self.instance,reply=_reply,reply_content=_reply_content,reply_author=_reply_author)
		else:
			raise s.exception(s.reason)

	async def send_message(self,channel_id:str,message:str="", embeds:List[Embed]=[],file:Union[File,None]=None,tts:bool=False,_reply:Union[Reply,None]=None) -> Message:
		from ... import HTTPStatus
		if message == embeds == file == None:
			raise Exception("Please inlcude either a message content, embed, or file.")
		if type(message) != str and message != None:
			raise Exception("Please pass in a 'str' object for the 'message' parameter.")
		if not type(embeds) == list:
			raise Exception("Please pass in a list of Embed objects for the 'embeds' parameter.")
		if type(file) != File and file != None:
			raise Exception("Please pass in a File object for the 'file' parameter.")
		if not type(tts) == bool:
			raise Exception("Please pass in a bool for the 'tts' parameter.")

		data = {'content': message,'tts': tts}
		if file and not file.empty:
			if type(file) != File:
					raise Exception("Please pass a \"File\" object to the \"file\" parameter.")
			data = {'content': message,'tts':tts}
			fields = file.fields
			fields['payload_json'] = (None,json.dumps(data))
			data = MultipartEncoder(fields=fields,boundary='----WebKitFormBoundary'+''.join(random.sample(string.ascii_letters+string.digits,16)))
			self.s.headers['Content-Type'] = data.content_type
		else:
			if len(embeds) > 0:
				e = []
				for embed in embeds:
					e.append(await prepareEmbed(embed))
				data['embeds'] = e
			data = json.dumps(data)
		r = self.s.post(f"{self.instance.base_url}channels/{channel_id}/messages",data=data)
		self.s.headers['Content-Type'] = 'application/json'
		s = HTTPStatus(r)
		if s.success:
			reply_content = None
			reply_author = None
			if _reply:
				reply_content = _reply.content
				reply_author = _reply.author
			reply:bool = bool(_reply)
			return Message(r.json(),self.instance,reply=reply,reply_content=reply_content,reply_author=reply_author)
		else:
			raise s.exception(s.reason)

	async def bulk_delete(self,channel_id:str,messages:List[Union[Message,str]]) -> None:
		from ... import HTTPStatus
		ids = []
		for message in messages:
				ids.append(message.id) if isinstance(message,Message) else ids.append(message)
		for id in ids:
			r = self.s.delete(url=f"{self.instance.base_url}channels/{channel_id}/messages/{id}") #bulk delete endpoint doesnt exist (absolute stupid)
			s = HTTPStatus(r)
			#if not s.success:
			#	raise s.exception(s.reason)

	async def pin_message(self,channel_id:str,id:str) -> None:
		from ... import HTTPStatus
		r = self.s.put(f"{self.instance.base_url}channels/{channel_id}/pins/{id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def unpin_message(self,channel_id:str,id:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(f"{self.instance.base_url}channels/{channel_id}/pins/{id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)



from .guild import PartialGuild,Channel
from .role import Role, Permissions
from .file import Icon

from typing import *

class Profile:
	def __init__(self,data,instance):
		self.instance = instance
		self.connected_accounts:list = data['connected_accounts']
		self.relationships:list[User] = [User(user,instance) for user in data['relationships']]
		self.mutual_guilds:list[PartialGuild] = [PartialGuild(guild,instance) for guild in data['mutual_guilds']]
		self.premium_since:str = data['premium_since']
		self.user:User = User(data['user'],instance)

	async def send_friend_request(self) -> None:
		return await self.instance.add_friend(self.user.id)

class Member:
	def __init__(self,data,guild_id,instance,roles=None):
		self.instance = instance
		self.guild_id:str = guild_id
		self.id:str = data['id']
		self.mention:str = f"<@{data['id']}>"
		self.nick:str = data['nick']
		self.roles:list[Role] = []
		if roles:
			for role in roles:
				if role.id in data['roles'] or role.id == guild_id:
					self.roles.append(role)
		else:
			for guild in instance.guilds:
				if guild.id == guild_id:
					for role in guild.roles:
						if role.id in data['roles'] or role.id == guild_id:
							self.roles.append(role)
		self.joined_at:str = data['joined_at']
		self.deaf:bool = data['deaf']
		self.mute:bool = data['mute']
		self.user:User = User(data['user'],instance,guild_id)
		self.permissions = Permissions()
		for role in self.roles:
			perms = role.permissions.get_dict()
			self.permissions.update(**perms)

	async def hasPermission(self,perm:str) -> bool:
		for role in self.roles:
			if await role.hasPermission(perm):
				return True
		for guild in self.instance.guilds:
			if guild.id == self.guild_id and guild.owner_id == self.id:
				return True
		return False

	async def hasChannelPermission(self,channel:Channel,permission:str) -> bool:
		if await self.hasPermission("administrator"):
			return True
		if type(channel) == "str": #in case someone does a thingy, also setting initial type to union will redline a bunch of stuff
			for guild in self.instance.guilds:
				if guild.id == self.guild_id:
					for c in guild.channels:
						if c.id == channel:
							channel = c
							break
					break
		for overwrite in channel.permission_overwrites:
			if overwrite.id == self.id:
				perm = await overwrite.hasPermission(permission)
				if perm is None:
					break
				elif perm:
					return True
				elif not perm:
					return False
		for role in self.roles:
			if await role.hasChannelPermission(channel,permission):
				return True
		return False

	async def canTarget(self,user:'Member') -> bool:
		highest = 0
		for role in self.roles:
			if highest < role.position:
				highest = role.position
		for guild in self.instance.guilds:
			if guild.id == self.guild_id:
				if guild.owner_id == self.id:
					return True
				elif guild.owner_id == user.id:
					return False
		target_highest = 0
		for role in user.roles:
			if target_highest < role.position:
				target_highest = role.position
		return highest > target_highest

	async def send_friend_request(self) -> None:
		return await self.instance.add_friend(self.id)

	def toDict(self):
		return {"id":self.id, "nick":self.nick, "roles":self.roles,"joined_at":self.joined_at, "deaf":self.deaf, "mute":self.mute,"user":self.user.toDict()}

	async def ban(self,reason:Union[str,None]=None,delete_message_days:int=0) -> None:
		return await self.instance.http.ban(self.guild_id,self.id,reason,delete_message_days)

	async def unban(self) -> None:
		return await self.instance.http.unban(self.guild_id,self.id)

	async def setRoles(self,roles:List[Union[Role,str]]) -> None:
		roles = [role.id if type(role) == Role else role for role in roles]
		roles.remove(self.guild_id)
		return await self.instance.http.update_guild_member(self.guild_id,self.id,roles=roles)

	async def removeRoles(self,roles:List[Union[Role,str]]):
		temp = [role.id for role in self.roles]
		for role in roles:
			if type(role) == Role:
				temp.remove(role.id)
			else:
				temp.remove(role)
		roles = temp
		roles.remove(self.guild_id)
		return await self.instance.http.update_guild_member(self.guild_id,self.id,roles=roles)

	async def addRoles(self,roles:List[Union[Role,str]]) -> None:
		temp = [role.id if type(role) == Role else role for role in self.roles]
		roles = [role.id if type(role) == Role else role for role in roles]
		for role in roles:
			temp.append(role)
		roles = temp
		roles.remove(self.guild_id)
		return await self.instance.http.update_guild_member(self.guild_id,self.id,roles=roles)

	async def kick(self,reason:Union[str,None]=None) -> None:
		return await self.instance.http.kick(self.guild_id,self.id,reason)

	async def setNick(self,nick:str) -> None:
		return await self.instance.http.update_guild_member(self.guild_id,self.id,nick=nick)

class Presence:
	def __init__(self,data,guild_id,instance):
		self.instance = instance
		self.guild_id:Union[str,None] = guild_id
		self.game:str = data.get('game')
		self.status:str = data['status']
		self.user:User = User(data['user'],instance,guild_id)

	def toDict(self):
		return {"game":self.game,"status":self.status,"user":self.user.toDict()}

class User:
	def __init__(self,data,instance,guild_id=None):
		self.instance = instance
		self.guild_id:Union[str,None] = guild_id
		self.id:str = data['id']
		self.username:str = data['username']
		self.discriminator:str = data['discriminator']
		self.mention:str = f"<@{data['id']}>"
		self.avatar:Icon = Icon(data,"avatar",self.instance,True)
		self.bot:bool = data.get('bot')

	async def send_friend_request(self) -> None:
		return await self.instance.add_friend(self.id)

	async def getProfile(self) -> Profile:
		return await self.instance.http.get_user_profile(self.id)

	async def getGuildMember(self,guild_id:Union[str,None]=None) -> Member:
		guild_id = guild_id if guild_id else self.guild_id
		if not guild_id:
			raise Exception("This user was not found in a guild.")
		for guild in self.instance.guilds:
			if guild.id == guild_id:
				for member in guild.members:
					if member.id == self.id:
						return member
		raise Exception("Client is not in a guild with the user!")

	def toDict(self):
		return {"id":self.id,"username":self.username,"discriminator":self.discriminator,"mention":self.mention,"avatar":self.avatar.url,"bot":self.bot}

	async def ban(self,reason:Union[str,None]=None,delete_message_days:int=0) -> None:
		if not self.guild_id:
			raise Exception("This user was not found in a guild.")
		return await self.instance.http.ban(self.guild_id,self.id,reason,delete_message_days)

	async def unban(self) -> None:
		if not self.guild_id:
			raise Exception("This user was not found in a guild.")
		return await self.instance.http.unban(self.guild_id,self.id)

	async def setRoles(self,roles:List[Union[Role,str]]) -> None:
		if not self.guild_id:
			raise Exception("This user was not found in a guild.")
		roles = [role.id if type(role) == Role else role for role in roles]
		roles.remove(self.guild_id)
		return await self.instance.http.update_guild_member(self.guild_id,self.id,roles=roles)

	async def removeRoles(self,roles:List[Union[Role,str]]) -> None:
		if not self.guild_id:
			raise Exception("This user was not found in a guild.")
		temp = await self.getGuildMember()
		temp = [role.id for role in temp.roles]
		for role in roles:
			if type(role) == Role:
				temp.remove(role.id)
			else:
				temp.remove(role)
		roles = temp
		roles.remove(self.guild_id)
		return await self.instance.http.update_guild_member(self.guild_id,self.id,roles=roles)

	async def addRoles(self,roles:List[Union[Role,str]]) -> None:
		if not self.guild_id:
			raise Exception("This user was not found in a guild.")
		temp = await self.getGuildMember()
		temp = [role.id for role in temp.roles]
		for role in roles:
			temp.append(role.id if type(role) == Role else role)
		roles = temp
		roles.remove(self.guild_id)
		return await self.instance.http.update_guild_member(self.guild_id,self.id,roles=roles)

	async def kick(self,reason:Union[str,None]=None) -> None:
		if not self.guild_id:
			raise Exception("This user was not found in a guild.")
		return await self.instance.http.kick(self.guild_id,self.id,reason)

	async def nick(self,nick:str) -> None:
		if not self.guild_id:
			raise Exception("This user was not found in a guild.")
		return await self.instance.http.update_guild_member(self.guild_id,self.id,nick=nick)



from .guild import Channel,Guild
from .user import User,Presence
from .file import Icon

from typing import *

class Ready:
	def __init__(self,data,instance):
		self.instance = instance
		self.version:int = data.get('v')
		self.user:Self = Self(data['user'],instance)
		self.private_channels:list[Channel] = [Channel(data,instance) for data in data['private_channels']]
		self.guilds_are_unavailable:bool = bool(data['guilds'][0].get('unavailable'))
		if not self.guilds_are_unavailable:
			self.guilds:list[Guild] = [Guild(guild,instance) for guild in data['guilds']]
		else:
			self.guild_ids:list[str] = [guild['id'] for guild in data['guilds']]
		if not data.get('users'):
		    data['users'] = []
		self.users:list[User] = [User(user,instance) for user in data['users']]
		self.presences:list[Presence] = [Presence(data,None,instance) for data in data['presences']]
		self.shard:list[int] = data.get('shard')
		self.session_id:str = data['session_id']
		self._trace:list[str] = data['_trace']
		self.relationships:list[Relationship] = [Relationship(relation,instance) for relation in data['relationships']]
		self.application:Union[Application,None] = None
		if data.get('application'):
			data['application']['discriminator'] = self.user.discriminator
			self.application = Application(data['application'],instance)

class Self(User):
	def __init__(self,data,instance):
		super().__init__(data,instance)
		self.email:str = data['email']
		self.verified:bool = data['verified']
		self.flags:int = data.get('flags')
		self.mfa_enabled:bool = data.get('mfa_enabled')
		self.premium:bool = data['premium']
		self.token:Union[str,None] = data.get('token') #PATCH /users/@me returns this

class Relationship(User):
	def __init__(self,data,instance):
		super().__init__(data['user'],instance)
		self.type:int = data['type']

class Application:
	def __init__(self,data,instance):
		self.instance = instance
		self.id:str = data['id']
		self.name:str = data['name']
		self.description:str = data['description']
		self.icon:Icon = Icon(data,"icon",instance,True)
		self.bot_public:bool = data['bot_public']
		self.bot_require_code_grant:bool = data['bot_require_code_grant']



from typing import *

class Aliases:
	def __init__(self):
		self.aliases = {}

	def add_aliases(self,aliases:List[str]):
		def decorator(func):
			for alias in aliases:
				self.aliases[alias] = func
			return func
		return decorator

class Commands:
	def __init__(self,prefix:str,aliases:Union[Aliases,None]=None):
		from ..main import Client
		self.instance:Client = None
		self.prefix = prefix
		self.aliases = aliases



from ...ready import Self
from ...file import File

from typing import *

import base64

class hSelf:
	def __init__(self,instance):
		self.instance = instance
		self.s = instance.s

	async def update_user(self,username:Union[str,None]=None,avatar:Union[File,None]=None) -> Self:
		from ...http import HTTPStatus
		if not username:
			username = self.instance.user.username
		data = {"username":username}
		if avatar:
			data["avatar"] = f"data:image/png;base64,{base64.b64encode(await avatar.get_file_data()).decode('utf-8')}"
		else:
			e = self.s.get(f"{self.instance.base_url}users/@me/").json()
			avdata = self.s.get(f"{self.instance.cdn}avatars/{e['id']}/{e['avatar']}.png").content
			data["avatar"] = f"data:image/png;base64,{base64.b64encode(avdata).decode('utf-8')}"
		r = self.s.patch(f"{self.instance.base_url}users/@me/",json=data)
		s = HTTPStatus(r)
		if s.success:
			return Self(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def add_friend(self,user_id:str) -> None:
		from ... import HTTPStatus
		r = self.s.put(f"{self.instance.base_url}users/@me/relationships/{user_id}",json={"type":False})
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def block(self,user_id:str) -> None:
		from ... import HTTPStatus
		r = self.s.put(f"{self.instance.base_url}users/@me/relationships/{user_id}",json={"type":2})
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def remove_relationship(self,user_id:str) -> None: # this is supposed to bhe unblock | uh sir ever heard of removing friends? | this was the request to unblock somone
		from ... import HTTPStatus
		r = self.s.delete(f"{self.instance.base_url}users/@me/relationships/{user_id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)


from ..embed import Embed, ImageEmbed

async def prepareEmbed(embed):
	prep_embed = {"title":embed.title,"description":embed.description,"color":embed.color,"url":embed.url,"timestamp":embed.timestamp,"fields":[]}
	if embed.image:
		prep_embed["image"] = {"url":embed.image.url,"width":embed.image.width,"height":embed.image.height}
	if embed.thumbnail:
		prep_embed["thumbnail"] = {"url":embed.thumbnail.url,"width":embed.thumbnail.width,"height":embed.thumbnail.height}
	if embed.footer:
		prep_embed["footer"] = {"text":embed.footer.text,"icon_url":embed.footer.icon_url}
	if embed.author:
		prep_embed["author"] = {"name":embed.author.name,"url":embed.author.url,"icon_url":embed.author.icon_url}
	for field in embed.fields:
		prep_embed['fields'].append({"name":field.name,"value":field.value,"inline":field.inline})
	return prep_embed

def makeEmbed(embeds):
	prep = []
	for embed in embeds:
		if embed['type'] == "image" or embed['type'] == "gifv":
			e = ImageEmbed(embed)
		else:
			e = Embed(embed.get('title'),embed.get('description'),embed.get('color'),embed.get('url'),embed.get('timestamp'))
		if embed.get('fields') and type(e) == Embed:
			for field in embed['fields']:
				e._addField(field.get('name'),field.get('value'),field.get('inline'))
		if embed.get('video'):
			e._addVideo(embed['video']['url'],embed['video']['width'],embed['video']['height'])
		if embed.get('provider'):
			e._addProvider(embed['provider'])
		if embed.get('image'):
			e._addImage(embed['image']['url'])
		if embed.get('thumbnail'):
			e._addThumbnail(embed['thumbnail']['url'])
		if embed.get('footer'):
			e._addFooter(embed['footer']['text'],embed['footer'].get('icon_url'))
		if embed.get('author'):
			e._addAuthor(embed['author']['name'],embed['author'].get('url'),embed['author'].get('icon_url'))
		prep.append(e)
	return prep


class Reply:
	def __init__(self,message):
		self.content = message.content
		self.author = message.author


from typing import *

class Role:
	def __init__(self,data,guild_id,instance):
		self.instance = instance
		self.guild_id:str = guild_id
		self.id:str = data['id']
		self.name:str = data['name']
		self.color:int = data['color']
		self.permissions:Permissions = Permissions(data['permissions'])
		self.mentionable:bool = data['mentionable']
		self.hoist:bool = data['hoist']
		self.position:int = data['position']

	async def hasPermission(self,permission) -> bool:
		return await self.permissions.hasPermission(permission)

	async def hasChannelPermission(self,channel,permission) -> bool:
		for overwrite in channel.permission_overwrites:
			if overwrite.id == self.id:
				perm = await overwrite.hasPermission(permission)
				if perm is None:
					break
				elif perm:
					return True
				elif not perm:
					return False
		return await self.hasPermission(permission)

	async def edit(self,name=None,permissions=None,color=None,hoist=None,mentionable=None) -> 'Role':
		return await self.instance.http.modify_role(self.guild_id,self.id,name,permissions,color,hoist,mentionable)

	async def updatePosition(self,position) -> List['Role']:
		return await self.instance.http.modify_role_position(self.guild_id,self.id,position)

	async def delete(self) -> None:
		return await self.instance.http.delete_role(self.guild_id,self.id)

	def toDict(self):
		return {"id":self.id, "name":self.name, "color":self.color, "permissions":self.permissions.value, "mentionable":self.mentionable, "hoist":self.hoist,"position":self.position}

class FakeRole: #this is really used to provide a placeholder for top role comparison
	def __init__(self,position):
		self.position = position


#stolen from dpy lol
class Permissions:
	"""Wraps up the Discord permission value.

	Supported operations:

	+-----------+------------------------------------------+
	| Operation |               Description                |
	+===========+==========================================+
	| x == y    | Checks if two permissions are equal.     |
	+-----------+------------------------------------------+
	| x != y    | Checks if two permissions are not equal. |
	+-----------+------------------------------------------+
	| x <= y    | Checks if a permission is a subset       |
	|           | of another permission.                   |
	+-----------+------------------------------------------+
	| x >= y    | Checks if a permission is a superset     |
	|           | of another permission.                   |
	+-----------+------------------------------------------+
	| x < y     | Checks if a permission is a strict       |
	|           | subset of another permission.            |
	+-----------+------------------------------------------+
	| x > y     | Checks if a permission is a strict       |
	|           | superset of another permission.          |
	+-----------+------------------------------------------+
	| hash(x)   | Return the permission's hash.            |
	+-----------+------------------------------------------+
	| iter(x)   | Returns an iterator of (perm, value)     |
	|           | pairs. This allows this class to be used |
	|           | as an iterable in e.g. set/list/dict     |
	|           | constructions.                           |
	+-----------+------------------------------------------+

	The properties provided are two way. You can set and retrieve individual bits using the properties as if they
	were regular bools. This allows you to edit permissions.

	Attributes
	-----------
	value
					The raw value. This value is a bit array field of a 32-bit integer
					representing the currently available permissions. You should query
					permissions via the properties rather than using this raw value.
	"""

	__slots__ = ('value',)
	def __init__(self, obj=0): #modified dpy to accept more objects
		if type(obj) == Role:
			obj.permissions.update()
			permissions = obj.permissions.value
		elif type(obj) == int:
			permissions = obj
		elif type(obj) == str:
			permissions = int(obj)
		elif not obj: #None
			permissions = 0 #creation of a fake role
		else:
			raise TypeError(f'Expected int, Role, str, or None, received {obj.__class__.__name__} instead.')

		self.value = permissions

	async def hasPermission(self,permission) -> bool: #the one function I (LG) actually entirely made
		if self.administrator:
			return True
		from .utils import Exceptions
		try:
			return getattr(self,permission)
		except AttributeError:
			raise Exceptions.InvalidPermission(f"{permission} is not a valid permission!")

	def __eq__(self, other):
		return isinstance(other, Permissions) and self.value == other.value

	def __ne__(self, other):
		return not self.__eq__(other)

	def __hash__(self):
		return hash(self.value)

	def __repr__(self):
		return '<Permissions value=%s>' % self.value

	def _perm_iterator(self):
		for attr in dir(self):
			# check if it's a property, because if so it's a permission
			is_property = isinstance(getattr(self.__class__, attr), property)
			if is_property:
				yield (attr, getattr(self, attr))

	def __iter__(self):
		return self._perm_iterator()

	def is_subset(self, other):
		"""Returns True if self has the same or fewer permissions as other."""
		if isinstance(other, Permissions):
			return (self.value & other.value) == self.value
		else:
			raise TypeError("cannot compare {} with {}".format(self.__class__.__name__, other.__class__name))

	def is_superset(self, other):
		"""Returns True if self has the same or more permissions as other."""
		if isinstance(other, Permissions):
			return (self.value | other.value) == self.value
		else:
			raise TypeError("cannot compare {} with {}".format(self.__class__.__name__, other.__class__name))

	def is_strict_subset(self, other):
		"""Returns True if the permissions on other are a strict subset of those on self."""
		return self.is_subset(other) and self != other

	def is_strict_superset(self, other):
		"""Returns True if the permissions on other are a strict superset of those on self."""
		return self.is_superset(other) and self != other

	__le__ = is_subset
	__ge__ = is_superset
	__lt__ = is_strict_subset
	__gt__ = is_strict_superset

	@classmethod
	def none(cls):
		"""A factory method that creates a :class:`Permissions` with all
		permissions set to False."""
		return cls(0)

	@classmethod
	def all(cls):
		"""A factory method that creates a :class:`Permissions` with all
		permissions set to True."""
		return cls(0b01111111111101111111110001111111)

	@classmethod
	def all_channel(cls):
		"""A :class:`Permissions` with all channel-specific permissions set to
		True and the guild-specific ones set to False. The guild-specific
		permissions are currently:

		- manage_guild
		- kick_members
		- ban_members
		- administrator
		- change_nicknames
		- manage_nicknames
		"""
		return cls(0b00110011111101111111110001010001)

	@classmethod
	def general(cls):
		"""A factory method that creates a :class:`Permissions` with all
		"General" permissions from the official Discord UI set to True."""
		return cls(0b01111100000000000000000000111111)

	@classmethod
	def text(cls):
		"""A factory method that creates a :class:`Permissions` with all
		"Text" permissions from the official Discord UI set to True."""
		return cls(0b00000000000001111111110001000000)

	@classmethod
	def voice(cls):
		"""A factory method that creates a :class:`Permissions` with all
		"Voice" permissions from the official Discord UI set to True."""
		return cls(0b00000011111100000000000000000000)
	
	def get_dict(self): #i also made this lol
		vals = {}
		for attr in dir(self):
			val = getattr(self.__class__,attr)
			if isinstance(attr,property) and val:
				vals[attr.__str__] = val
		return vals

	def update(self, **kwargs):
		"""Bulk updates this permission object.

		Allows you to set multiple attributes by using keyword
		arguments. The names must be equivalent to the properties
		listed. Extraneous key/value pairs will be silently ignored.

		Parameters
		------------
		\*\*kwargs
						A list of key/value pairs to bulk update permissions with.
		"""
		for key, value in kwargs.items():
			try:
				is_property = isinstance(getattr(self.__class__, key), property)
			except AttributeError:
				continue

			if is_property:
				setattr(self, key, value)

	def _bit(self, index):
		return bool((self.value >> index) & 1)

	def _set(self, index, value):
		if value == True:
			self.value |= (1 << index)
		elif value == False:
			self.value &= ~(1 << index)
		else:
			raise TypeError('Value to set for Permissions must be a bool.')

	def handle_overwrite(self, allow, deny):
		# Basically this is what's happening here.
		# We have an original bit array, e.g. 1010
		# Then we have another bit array that is 'denied', e.g. 1111
		# And then we have the last one which is 'allowed', e.g. 0101
		# We want original OP denied to end up resulting in
		# whatever is in denied to be set to 0.
		# So 1010 OP 1111 -> 0000
		# Then we take this value and look at the allowed values.
		# And whatever is allowed is set to 1.
		# So 0000 OP2 0101 -> 0101
		# The OP is base  & ~denied.
		# The OP2 is base | allowed.
		self.value = (self.value & ~deny) | allow

	@property
	def create_instant_invite(self):
		"""Returns True if the user can create instant invites."""
		return self._bit(0)

	@create_instant_invite.setter
	def create_instant_invite(self, value):
		self._set(0, value)

	@property
	def kick_members(self):
		"""Returns True if the user can kick users from the guild."""
		return self._bit(1)

	@kick_members.setter
	def kick_members(self, value):
		self._set(1, value)

	@property
	def ban_members(self):
		"""Returns True if a user can ban users from the guild."""
		return self._bit(2)

	@ban_members.setter
	def ban_members(self, value):
		self._set(2, value)

	@property
	def administrator(self):
		"""Returns True if a user is an administrator. This role overrides all other permissions.

		This also bypasses all channel-specific overrides.
		"""
		return self._bit(3)

	@administrator.setter
	def administrator(self, value):
		self._set(3, value)

	@property
	def manage_channels(self):
		"""Returns True if a user can edit, delete, or create channels in the guild.

		This also corresponds to the "manage channel" channel-specific override."""
		return self._bit(4)

	@manage_channels.setter
	def manage_channels(self, value):
		self._set(4, value)

	@property
	def manage_guild(self):
		"""Returns True if a user can edit guild properties."""
		return self._bit(5)

	@manage_guild.setter
	def manage_guild(self, value):
		self._set(5, value)

	@property
	def add_reactions(self):
		"""Returns True if a user can add reactions to messages."""
		return self._bit(6)

	@add_reactions.setter
	def add_reactions(self, value):
		self._set(6, value)

	# 4 unused

	@property
	def read_messages(self):
		"""Returns True if a user can read messages from all or specific text channels."""
		return self._bit(10)

	@read_messages.setter
	def read_messages(self, value):
		self._set(10, value)

	@property
	def send_messages(self):
		"""Returns True if a user can send messages from all or specific text channels."""
		return self._bit(11)

	@send_messages.setter
	def send_messages(self, value):
		self._set(11, value)

	@property
	def send_tts_messages(self):
		"""Returns True if a user can send TTS messages from all or specific text channels."""
		return self._bit(12)

	@send_tts_messages.setter
	def send_tts_messages(self, value):
		self._set(12, value)

	@property
	def manage_messages(self):
		"""Returns True if a user can delete messages from a text channel. Note that there are currently no ways to edit other people's messages."""
		return self._bit(13)

	@manage_messages.setter
	def manage_messages(self, value):
		self._set(13, value)

	@property
	def embed_links(self):
		"""Returns True if a user's messages will automatically be embedded by Discord."""
		return self._bit(14)

	@embed_links.setter
	def embed_links(self, value):
		self._set(14, value)

	@property
	def attach_files(self):
		"""Returns True if a user can send files in their messages."""
		return self._bit(15)

	@attach_files.setter
	def attach_files(self, value):
		self._set(15, value)

	@property
	def read_message_history(self):
		"""Returns True if a user can read a text channel's previous messages."""
		return self._bit(16)

	@read_message_history.setter
	def read_message_history(self, value):
		self._set(16, value)

	@property
	def mention_everyone(self):
		"""Returns True if a user's @everyone will mention everyone in the text channel."""
		return self._bit(17)

	@mention_everyone.setter
	def mention_everyone(self, value):
		self._set(17, value)

	@property
	def external_emojis(self):
		"""Returns True if a user can use emojis from other guilds."""
		return self._bit(18)

	@external_emojis.setter
	def external_emojis(self, value):
		self._set(18, value)

	# 1 unused

	@property
	def connect(self):
		"""Returns True if a user can connect to a voice channel."""
		return self._bit(20)

	@connect.setter
	def connect(self, value):
		self._set(20, value)

	@property
	def speak(self):
		"""Returns True if a user can speak in a voice channel."""
		return self._bit(21)

	@speak.setter
	def speak(self, value):
		self._set(21, value)

	@property
	def mute_members(self):
		"""Returns True if a user can mute other users."""
		return self._bit(22)

	@mute_members.setter
	def mute_members(self, value):
		self._set(22, value)

	@property
	def deafen_members(self):
		"""Returns True if a user can deafen other users."""
		return self._bit(23)

	@deafen_members.setter
	def deafen_members(self, value):
		self._set(23, value)

	@property
	def move_members(self):
		"""Returns True if a user can move users between other voice channels."""
		return self._bit(24)

	@move_members.setter
	def move_members(self, value):
		self._set(24, value)

	@property
	def use_voice_activation(self):
		"""Returns True if a user can use voice activation in voice channels."""
		return self._bit(25)

	@use_voice_activation.setter
	def use_voice_activation(self, value):
		self._set(25, value)

	@property
	def change_nickname(self):
		"""Returns True if a user can change their nickname in the guild."""
		return self._bit(26)

	@change_nickname.setter
	def change_nickname(self, value):
		self._set(26, value)

	@property
	def manage_nicknames(self):
		"""Returns True if a user can change other user's nickname in the guild."""
		return self._bit(27)

	@manage_nicknames.setter
	def manage_nicknames(self, value):
		self._set(27, value)

	@property
	def manage_roles(self):
		"""Returns True if a user can create or edit roles less than their role's position.

		This also corresponds to the "manage permissions" channel-specific override.
		"""
		return self._bit(28)

	@manage_roles.setter
	def manage_roles(self, value):
		self._set(28, value)

	@property
	def manage_webhooks(self):
		"""Returns True if a user can create, edit, or delete webhooks."""
		return self._bit(29)

	@manage_webhooks.setter
	def manage_webhooks(self, value):
		self._set(29, value)

	@property
	def manage_emojis(self):
		"""Returns True if a user can create, edit, or delete emojis."""
		return self._bit(30)

	@manage_emojis.setter
	def manage_emojis(self, value):
		self._set(30, value)



import random

class Enums:
	class Presence:
		online = "online"
		idle = "idle"
		do_not_disturb = "dnd"
		invisible = "invisible"

	class Status:
		playing = 1
		listening_to = 2
		watching = 3

	class Channel:
		text = 0
		dm = 1
		voice = 2
		gc = 3

	class DefaultAvatars:
		all = ("https://hummus.sys42.net/assets/6debd47ed13483642cf09e832ed0bc1b.png","https://hummus.sys42.net/assets/322c936a8c8be1b803cd94861bdfa868.png","https://hummus.sys42.net/assets/dd4dbc0016779df1378e7812eabaa04d.png","https://hummus.sys42.net/assets/0e291f67c9274a1abdddeb3fd919cbaa.png","https://hummus.sys42.net/assets/1cbd08c76f8af6dddce02c5138971129.png") #cause ziad is stupid and didnt put the default avatars under /embeds/avatars
		blue = "https://hummus.sys42.net/assets/6debd47ed13483642cf09e832ed0bc1b.png"
		gray = "https://hummus.sys42.net/assets/322c936a8c8be1b803cd94861bdfa868.png"
		green = "https://hummus.sys42.net/assets/dd4dbc0016779df1378e7812eabaa04d.png"
		yellow = "https://hummus.sys42.net/assets/0e291f67c9274a1abdddeb3fd919cbaa.png"
		red = "https://hummus.sys42.net/assets/1cbd08c76f8af6dddce02c5138971129.png"

	class Permissions:
		add_reactions = "add_reactions"
		administrator = "administrator"
		attach_files = "attach_files"
		ban_members = "ban_members"
		change_nickname = "change_nickname"
		connect = "connect"
		create_instant_invite = "create_instant_invite"
		deafen_members = "deafen_members"
		embed_links = "embed_links"
		external_emojis = "external_emojis"
		kick_members = "kick_members"
		manage_channels = "manage_channels"
		manage_emojis = "manage_emojis"
		manage_guild = "manage_guild"
		manage_messages = "manage_messages"
		manage_nicknames = "manage_nicknames"
		manage_roles = "manage_roles"
		manage_webhooks = "manage_webhooks"
		mention_everyone = "mention_everyone"
		move_members = "move_members"
		mute_members = "mute_members"
		read_message_history = "read_message_history"
		read_messages = "read_messages"
		send_messages = "send_messages"
		send_tts_messages = "send_tts_messages"
		speak = "speak"
		use_voice_activation = "use_voice_activation"

	#copied from fossbotpy lol
	class Colors:
		default = 0 #just another name for black ig
		aqua = 0x1ABC9C
		dark_aqua = 0x11806A
		green = 0x2ECC71
		dark_green = 0x1F8B4C
		blue = 0x3498DB
		dark_blue = 0x206694
		purple = 0x9B59B6
		dark_purple = 0x71368A
		magenta = 0xE91E63
		dark_magenta = 0xAD1457
		gold = 0xF1C40F
		dark_gold = 0xC27C0E
		orange = 0xE67E22
		dark_orange = 0xA84300
		red = 0xE74c3C
		dark_red = 0x992D22
		light_gray = 0xBCC0C0
		gray = 0x95A5A6
		dark_gray = 0x979C9F
		darker_gray = 0x7F8C8D
		og_blurple = 0x7289DA
		ruined_blurple = 0x586AEA
		blurple = 0x5865F2
		greyple = 0x99AAB5
		dark_them = 0x36393F
		not_quite_black = 0x23272A
		dark_but_not_black = 0x2C2F33
		white = 0xFFFFFE #thx dolfies for noting that 0xFFFFFF doesn't work
		fuchsia = 0xEB459E
		yellow = 0xFEE75C
		navy = 0x34495E
		dark_navy = 0x2C3E50

		@staticmethod
		def get_random_color():
			return random.randint(0x000000, 0xFFFFFF)

		@staticmethod
		def get_byte(value, byte):
			return (value >> (8*byte)) & 0xFF

		@staticmethod
		def from_rgb(*args):
			newArgs = [0,0,0]
			if type(args[0]) in (list, tuple):
				newArgs[0],newArgs[1],newArgs[2] = args[0]
			else:
				newArgs = list(args)
			return (newArgs[0]<<16) + (newArgs[1]<<8) + newArgs[2]

		@staticmethod
		def to_rgb(c):
			return (Enums.Colors.get_byte(c,2), Enums.Colors.get_byte(c,1), Enums.Colors.get_byte(c,0))

		@staticmethod
		def get(*args): #accepts decimal, hex, rgb
			#args -> c
			if len(args) == 3:
				c = args
			elif len(args) == 1:
				c = args[0]
			else:
				raise ValueError("Expected either decimal, hex, or rgb input.")
			#parse c (color input)
			if isinstance(c, tuple) or isinstance(c, list):
				result = Enums.Colors.from_rgb(c)
			elif isinstance(c, str):
				if c.startswith("0x"):
					result = int(c, 0)
				else:
					c = c.lower()
					if c == "random":
						result = Enums.Colors.get_random_color()
					else:
						key = c.replace("grey", "gray")
						if key in dir(Enums.Colors):
							result = getattr(Enums.Colors,key)
						else:
							result = int(c, 16)
			else:
				result = c
			return result


#welcome to stupidity central
import websockets
from websockets.sync.client import connect
import _thread as thread
import fake_useragent
import traceback
import requests
import inspect
import asyncio
import json
import time

from .utils import Commands, CustomStatus, ClientSettings
from .utils import Context,CustomizableExceptions
from .user import User, Member, Presence
from .guild import PartialGuild, Guild
from .guild import Channel, Emoji
from .message import Message
from .ready import Ready
from .role import Role
from .http import HTTP

from datetime import datetime
from typing import *

ua = fake_useragent.UserAgent(browsers=['chrome', "firefox", "opera", "safari", "edge", "internet explorer"])
agent = ua.random


async def splitArgs(function,message,instance):
	func_args = list(inspect.signature(function).parameters.values())[1:]
	custom_kwargs = []

	args = message.content.replace('“','"').replace('”','"').replace("‘","\"").replace("’","\"").split() #filtering

	args.pop(0)
	if not args:
		return []

	refined = []
	in_quotes = False
	current_arg = ""

	for arg in args:
		if in_quotes:
			current_arg += " " + arg
			if arg.endswith("\""):
				in_quotes = False
				refined.append(current_arg.strip("\""))
				current_arg = ""
			continue

		if arg.startswith("\"") and not arg.endswith("\"") and instance.settings.split_by_quotations:
			in_quotes = True
			current_arg = arg.strip("\"")
		else:
			refined.append(arg)

	role_mention_idx = 0
	mention_idx = 0
	idx = 0
	mentioned = dict()
	for arg_name, arg in zip(func_args, refined):
		if idx >= len(refined):
			break

		if arg_name.annotation == str:
			custom_kwargs.append(str(arg))
		elif arg_name.annotation == int:
			try:
				custom_kwargs.append(int(arg))
			except Exception:
				await instance.custom_exceptions.onInvalidInteger(arg,message)
		elif arg_name.annotation == bool:
			if arg.lower() == "true":
				custom_kwargs.append(True)
			elif arg.lower() == "false":
				custom_kwargs.append(False)
			else:
				await instance.custom_exceptions.onInvalidBoolean(arg,message)
		elif arg_name.annotation == Role:
			clean = arg.replace("<@&","").replace(">","")
			try:
			    if not (arg.startswith("<@&") and arg.endswith(">")) and not (clean in [str(r.id) for r in message.mention_roles] or mentioned.get(clean)):
			        raise Exception
			    if mentioned.get(clean):
			        custom_kwargs.append(mentioned[clean])
			    else:
			        custom_kwargs.append(message.mention_roles[role_mention_idx])
			        mentioned[clean] = message.mention_roles[role_mention_idx]
			        role_mention_idx += 1
			except Exception as e:
			    await instance.custom_exceptions.onInvalidRole(arg,message)
		elif arg_name.annotation == User:
			clean = arg.replace("<@","").replace(">","").replace("!","")
			if not arg.startswith("<@") and not arg.endswith(">") and not (clean in [str(m.id) for m in message.mentions] or mentioned.get(clean)):
				return await instance.custom_exceptions.onInvalidMention(arg,message)
			if mentioned.get(clean):
				custom_kwargs.append(mentioned[clean])
			else:
				m = message.mentions[mention_idx]
				custom_kwargs.append(m)
				mentioned[clean] = m
				mention_idx += 1
		elif arg_name.annotation == Member:
			clean = arg.replace("<@","").replace(">","").replace("!","")
			try:
				if not arg.startswith("<@") and not arg.endswith(">") and not (clean in [str(m.id) for m in message.mentions] or mentioned.get(clean)):
					raise Exception #User.getGuildMember() can also raise an exception that should be caught, so we do not return here
				if mentioned.get(clean):
					custom_kwargs.append(mentioned[clean])
				else:
					m = await message.mentions[mention_idx].getGuildMember()
					custom_kwargs.append(m)
					mentioned[clean] = m
					mention_idx += 1
			except Exception:
				await instance.custom_exceptions.onInvalidMention(arg,message)
		elif arg_name.annotation == Channel:
			found = False
			for channel in message.guild.channels:
				if f"<#{channel.id}>" == arg:
					found = True
					custom_kwargs.append(channel)
					break
			if not found:
				return await instance.custom_exceptions.onInvalidChannel(arg,message)
		else:
			custom_kwargs.append(arg)

		idx += 1
	return custom_kwargs


default_exceptions = CustomizableExceptions()
default_settings = ClientSettings()
default_status = CustomStatus(type=0)

class Client:
	def __init__(self,token,commands:Union[Commands,None]=None,url="https://hummus.sys42.net/api/v6/",cdn=None,custom_status:CustomStatus=default_status,exception_handler:CustomizableExceptions=default_exceptions,settings:ClientSettings=default_settings):
		self.websocket = ""
		self.ping = None
		self.state:Ready = None #type:ignore
		self.custom_exceptions:CustomizableExceptions = exception_handler
		self.commands:Union[Commands,None] = commands
		self.settings:ClientSettings = settings
		if isinstance(commands,Commands):
			self.prefix = commands.prefix
			self.commands.instance = self #type:ignore
		elif commands is None:
			self.prefix = None
		else:
		    raise TypeError("Commands must be of type Commands.")
		self.s = requests.session()
		self.s.headers = {"Authorization":token,"Content-Type":"application/json","User-Agent":agent}
		self.http:HTTP = HTTP(self)
		if not url.endswith("/"):
			url = url + "/"
		e = requests.get(url+"gateway",headers={"User-Agent":agent})
		if e.json().get('url'):
			self.websocket = e.json()['url']+"?api=v6&encoding=json"
			if not self.websocket.startswith("wss://"):
				self.websocket = "wss://" + self.websocket
		else:
			raise Exception("Please tell the website owner to add '/gateway' as an API endpoint.")
		if cdn:
			self.cdn = cdn
		else:
			cdn = url.split(".")
			cdn[0] = cdn[0]+"-cdn"
			cdn = ".".join(cdn)
			self.cdn = cdn.replace("api/","").replace("v6/","")
		if not self.cdn.endswith("/"):
			self.cdn = cdn + "/"
		self.status = custom_status._toJson()
		self.token = token
		self.base_url = url

	def __log(self,message): #this will likely have file logging functionality at some point
		if self.settings.logging:
			print(message)

	async def run(self):
		reconnect = False
		last_event = None
		session = ""
		seq = None
		self.guilds:list[Guild] = []
		while True:
			try:
				with connect(self.websocket,additional_headers=self.s.headers,user_agent_header=ua.random,max_size=None) as websocket:
					self.connection = websocket
					self.__log("\033[32mrestarting...\033[0m")
					if reconnect:
						self.__log(f"\033[32mResuming session with sequence ID {seq}\033[0m")
						websocket.send(json.dumps({"d":{"token":self.token,"session_id":session,"seq":seq},"op":6}))
					else:
						websocket.send(json.dumps({'op': 2,'d': {'token':self.token,'presence': self.status}}))

					while True:
						event = json.loads(websocket.recv())
						if reconnect:
						    print(event)
						if event.get('s'):
						    seq = event['s']
						if not event.get('t'):
						    event['t'] = None

						if event['t']:
						    last_event = event

						if event['t'] == "READY":
							session = event['d']['session_id']
							self.__log("\033[32mready, recieving login guild data\033[0m")
							self.state = Ready(event['d'],self)
							self.user = self.state.user
							self.__start_time = datetime.now()
							websocket.send(json.dumps({"op":1}))
							if not self.state.guilds_are_unavailable:
								self.guilds = self.state.guilds
							thread.start_new_thread(asyncio.run,(self.on_ready(self.state),))

						if event['op'] == 1:
							self.__start_time = datetime.now()
							websocket.send(json.dumps({"op":1}))

						if event['op'] == 11:
							passed = datetime.now() - self.__start_time
							self.ping = f"{passed.total_seconds()*1000} ms"

						if event['t'] == "CHANNEL_PINS_UPDATE":
							thread.start_new_thread(asyncio.run,(self.channel_pins_update(event['d']['channel_id'],event['d'].get('last_pin_timestamp')),))

						if event['t'] == "GUILD_BAN_ADD":
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									thread.start_new_thread(asyncio.run,(self.on_guild_ban_add(guild,User(event['d']['user'],self,event['d']['guild_id'])),))
									break

						if event['t'] == "GUILD_BAN_REMOVE":
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									thread.start_new_thread(asyncio.run,(self.on_guild_ban_remove(guild,User(event['d']['user'],self,event['d']['guild_id'])),))
									break

						if event['t'] == "TYPING_START":
							member = event['d'].get('member')
							channel = event['d']['channel_id']
							if member:
								is_guild = True
								member = Member(member,event['d']['guild_id'],self)
								for guild in self.guilds:
									if guild.id == event['d']['guild_id']:
										for c in guild.channels:
											if c.id == event['d']['channel_id']:
												channel = c
												break
										break
							else:
								is_guild = False
								member = event['d']['user_id']
								for dm in self.state.private_channels:
									if dm.id == event['d']['channel_id']:
										channel = dm
										for user in dm.recipients:
											if user.id == member:
												member = user
												break
										break
							thread.start_new_thread(asyncio.run,(self.on_typing_start(is_guild,channel,member,event['d']['timestamp']),))

						if event['t'] == "GUILD_MEMBER_UPDATE":
							guild_idx = 0
							member_idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									for member in guild.members:
										if member.id == event['d']['id']:
											new_member = Member(event['d'],event['d']['guild_id'],self)
											self.guilds[guild_idx].members[member_idx] = new_member
											thread.start_new_thread(asyncio.run,(self.on_guild_member_update(member,new_member),))
											break
										member_idx += 1
									break
								guild_idx += 1

						if event['t'] == "GUILD_UPDATE":
							guild_idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['id']:
									updated_guild = PartialGuild(event['d'],self)
									self.guilds[guild_idx]._update_data(updated_guild)
									thread.start_new_thread(asyncio.run,(self.on_guild_update(guild,updated_guild),))
									break
								guild_idx += 1

						if event['t'] == "GUILD_EMOJIS_UPDATE":
							guild_idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									before = guild.emojis
									after = [Emoji(emoji,self) for emoji in event['d']['emojis']]
									self.guilds[guild_idx].emojis = after
									thread.start_new_thread(asyncio.run,(self.on_guild_emojis_update(guild,before,after),))
									break
								guild_idx += 1

						if event['t'] == "GUILD_ROLE_UPDATE":
							guild_idx = 0
							role_idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									for role in guild.roles:
										if role.id == event['d']['role']['id']:
											new_role = Role(event['d']['role'],event['d']['guild_id'],self)
											self.guilds[guild_idx].roles[role_idx] = new_role
											thread.start_new_thread(asyncio.run,(self.on_guild_role_update(guild,role,new_role),))
											break
										role_idx += 1
									break
								guild_idx += 1

						if event['t'] == "GUILD_ROLE_CREATE":
							guild_idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									new_role = Role(event['d']['role'],event['d']['guild_id'],self)
									self.guilds[guild_idx].roles.append(new_role)
									thread.start_new_thread(asyncio.run,(self.on_guild_role_create(guild,new_role),))
									break
								guild_idx += 1

						if event['t'] == "GUILD_ROLE_DELETE":
							guild_idx = 0
							role_idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									for role in guild.roles:
										if role.id == event['d']['role_id']:
											deleted_role = role
											self.guilds[guild_idx].roles.pop(role_idx)
											thread.start_new_thread(asyncio.run,(self.on_guild_role_delete(guild,deleted_role),))
											break
										role_idx += 1
									break
								guild_idx += 1

						if event['t'] == "CHANNEL_CREATE":
							for dm in self.state.private_channels:
								if dm.id == event['d']['id']:
									channel = Channel(event['d'],self)
									self.state.private_channels.append(channel)
									thread.start_new_thread(asyncio.run,(self.on_channel_create(None,channel),))
							guild_idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									channel = Channel(event['d'],self)
									self.guilds[guild_idx].channels.append(channel)
									thread.start_new_thread(asyncio.run,(self.on_channel_create(guild,channel),))
									break
								guild_idx += 1

						if event['t'] == "CHANNEL_UPDATE":
							guild_idx = 0
							channel_idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									for channel in guild.channels:
										if channel.id == event['d']['id']:
											updated_channel = Channel(event['d'],self)
											self.guilds[guild_idx].channels[channel_idx] = updated_channel
											thread.start_new_thread(asyncio.run,(self.on_channel_update(guild,channel,updated_channel),))
											break
										channel_idx += 1
									break
								guild_idx += 1

						if event['t'] == "CHANNEL_DELETE":
							channel_idx = 0
							if not event['d'].get('guild_id'):
								for channel in self.state.private_channels:
									if channel.id == event['d']['id']:
										self.state.private_channels.pop(channel_idx)
										thread.start_new_thread(asyncio.run,(self.on_channel_delete(None,channel),))
										break
									channel_idx += 1
							guild_idx = 0
							for guild in self.guilds:
								if guild.id == event['d'].get('guild_id'):
									for channel in guild.channels:
										if channel.id == event['d']['id']:
											self.guilds[guild_idx].channels.pop(channel_idx)
											thread.start_new_thread(asyncio.run,(self.on_channel_delete(guild,channel),))
											break
										channel_idx += 1
									break
								guild_idx += 1

						if event['t'] == "PRESENCE_UPDATE":
							if event['d'].get('guild_id'):
								guild_idx = 0
								presence_idx = 0
								for guild in self.guilds:
									if guild.id == event['d']['guild_id']:
										for presence in guild.presences:
											if presence.user.id == event['d']['user']['id']:
												new_presence = Presence(event['d'],event['d']['guild_id'],self)
												self.guilds[guild_idx].presences[presence_idx] = new_presence
												thread.start_new_thread(asyncio.run,(self.on_presence_update(presence,new_presence),))
												break
											presence_idx += 1
										break
									guild_idx += 1
							else:
								presence_idx = 0
								for presence in self.state.presences:
									if presence.user.id == event['d']['user']['id']:
										self.state.presences[presence_idx] = Presence(event['d'],None,self)
										thread.start_new_thread(asyncio.run,(self.on_presence_update(presence,self.state.presences[presence_idx]),))

						if event['t'] == "GUILD_CREATE":
							guild = Guild(event['d'],self)
							self.guilds.append(guild)
							thread.start_new_thread(asyncio.run,(self.on_guild_create(guild),))

						if event['t'] == "GUILD_DELETE":
							idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['id']:
									self.guilds.pop(idx)
									thread.start_new_thread(asyncio.run,(self.on_guild_delete(guild),))
									break
								idx += 1

						if event['t'] == "GUILD_MEMBER_ADD":
							member = Member(event['d'],event['d']['guild_id'],self)
							idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									self.guilds[idx].members.append(member)
									thread.start_new_thread(asyncio.run,(self.on_guild_member_add(guild,member),))
									break
								idx += 1

						if event['t'] == "GUILD_MEMBER_REMOVE":
							member = User(event['d']['user'],self)
							idx = 0
							for guild in self.guilds:
								if guild.id == event['d']['guild_id']:
									midx = 0
									for member in self.guilds[idx].members:
										if member.id == event['d']['user']['id']:
											self.guilds[idx].members.pop(midx)
											thread.start_new_thread(asyncio.run,(self.on_guild_member_remove(guild,member),))
											break
										midx += 1
									break
								idx += 1

						if event['t'] == "MESSAGE_DELETE":
							guild = None
							if event['d'].get('guild_id'):
								guild_idx = 0
								for g in self.guilds:
									if g.id == event['d']['guild_id']:
										guild = g
										break
								thread.start_new_thread(asyncio.run,(self.on_message_delete(event['d']['id'],event['d']['channel_id'],guild),))

						if event['t'] == "MESSAGE_UPDATE":
							thread.start_new_thread(asyncio.run,(self.on_message_update(Message(event['d'],self)),))

						if event['t'] == "MESSAGE_CREATE":
							message = Message(event['d'],self)
							can_respond = True
							if not self.settings.reply_to_self and message.author.id == self.user.id:
								can_respond = False
							if not self.settings.reply_to_bots and message.author.bot:
								can_respond = False
							if can_respond or not self.settings.apply_to_events:
							    try:
							        await self.on_message_create(message)
							    except Exception as e:
							        print(f"\033[31mIgnoring exception: {e}\nTraceback:\033[0m")
							        traceback.print_exc()

							prefix = await self.get_prefix(message)
							if self.commands and prefix and event['d']['content'].startswith(prefix): #problem with uptime
								context = Context(message,self)
								func = message.content[len(prefix):]
								args = func.split()
								can_respond = True
								if len(args) < 1:
									can_respond = False
								if not self.settings.reply_to_self and message.author.id == self.user.id:
									can_respond = False
								if not self.settings.reply_to_bots and message.author.bot:
									can_respond = False
								if can_respond:
									try:
										command = getattr(self.commands, args[0])
										if callable(command) and not "__" in args[0]:
											try:
												command_args = await splitArgs(command,message,self)
												try:
													thread.start_new_thread(asyncio.run,(command(context,*command_args),))
												except TypeError as e:
													if "required positional argument" in str(e):
														args = str(e).split(": '")[1].replace("'","").replace(",","").replace("and","").split()
														await self.custom_exceptions.onMissingArguments(args,context)
											except Exception as ex:
												print(f"\033[31mIgnoring exception: {ex}\nTraceback:\033[0m")
												traceback.print_exc()
									except AttributeError as e:
										if self.commands.aliases and self.commands.aliases.aliases.get(args[0]):
											command = self.commands.aliases.aliases[args[0]]
											command = getattr(self.commands, command.__name__)
											try:
												command_args = await splitArgs(command,message,self)
												try:
													thread.start_new_thread(asyncio.run,(command(context,*command_args),))
												except TypeError as e:
													if "required positional argument" in str(e):
														args = str(e).split(": '")[1].replace("'","").replace(",","").replace("and","").split()
														await self.custom_exceptions.onMissingArguments(args,context)
											except Exception as e:
												print(f"\033[31mIgnoring exception: {e}\nTraceback:\033[0m")
										else:
											if args[0] in str(e):
												print(f"\033[31mCommand '{args[0]}' not found as a command or alias.\033[0m")
			except websockets.exceptions.ConnectionClosed as e:
				if e.code == 4004:
					print("\033[31mRecieved invalid token from server, exiting.\033[0m")
					quit()
			except websockets.exceptions.ConnectionClosedOK as e:
			    print(f"heartbeat expiry close, {e}\nLast event: {last_event}\nAt: {datetime.now()}")
			    traceback.print_exc()
			    reconnect = True
			except Exception as e:
				print(f"\033[31mDisconnected with error {e}, reconnecting...\033[0m")
				traceback.print_exc()
				reconnect = True
				time.sleep(5)

	async def get_prefix(self,message):
		return self.prefix

	#events
	async def on_ready(self, bot):
		pass

	async def channel_pins_update(self, channel_id, last_pin_timestamp):
		pass

	async def on_message_create(self, message):
		pass

	async def on_message_delete(self, message_id, channel_id, guild): #channel id in case this is in a dm or gc and we cannot get the channel object from the corresponding guild object, guild will be None if so
		pass

	async def on_message_update(self, message):
		pass

	async def on_presence_update(self, before, presence):
		pass

	async def on_guild_ban_add(self, guild, user):
		pass

	async def on_guild_ban_remove(self, guild, user):
		pass

	async def on_guild_member_add(self, guild, member):
		pass

	async def on_guild_member_remove(self, guild, member): #unlike other events, a member object actually gets passed through! discord is not being obscure with info for once!?!?!?
		pass

	async def on_guild_member_update(self, before, after):
		pass

	async def on_guild_create(self, guild):
		pass

	async def on_guild_delete(self, guild):
		pass

	async def on_guild_update(self, before, after):
		pass

	async def on_guild_role_create(self, guild, role):
		pass

	async def on_guild_role_delete(self, guild, role):
		pass

	async def on_guild_role_update(self, guild, before, after):
		pass

	async def on_guild_emojis_update(self, guild, before, after):
		pass

	async def on_typing_start(self, is_guild, channel, user, when):
		pass

	async def on_channel_create(self, guild, channel):
		pass

	async def on_channel_delete(self, guild, channel):
		pass

	async def on_channel_update(self, guild, before, after):
		pass



from .role import Role, Permissions
from .file import File, Icon
from .message import Message
from .embed import Embed

from typing import *
import contextlib
import asyncio

class PermissionOverwrite:
	def __init__(self,data):
		from .user import User, Member
		if type(data) == dict:
			self.id:str = data['id']
			self.type:str = data['type']
			self.allow:Permissions = Permissions(data['allow'])
			self.deny:Permissions = Permissions(data['deny'])
		if type(data) == Role:
			self.id:str = data.id
			self.type:str = "role"
			self.allow:Permissions = data.permissions
			self.deny:Permissions = Permissions(0)
		if type(data) == Member or type(data) == User:
			self.id:str = data.id
			self.type:str = "member"
			self.allow:Permissions = Permissions(0)
			self.deny:Permissions = Permissions(0)

	async def hasPermission(self,perm:str) -> Union[bool,None]:
		allow = await self.allow.hasPermission(perm)
		deny = await self.deny.hasPermission(perm)
		if allow == deny == False:
			return None
		elif allow == True:
			return True
		elif deny == True:
			return False

	def toDict(self):
		self.allow.update()
		self.deny.update()
		return {"id":self.id,"type":self.type,"allow":self.allow.value,"deny":self.deny.value}

class InvitePartialGuild:
	def __init__(self,data,instance):
		self.instance = instance
		self.id:str = data['id']
		self.name:str = data['name']
		self.splash:Union[str,None] = data.get('splash')
		self.icon:Icon = Icon(data,"icon",instance)
		self.owner_id:str = data['owner_id']
		self.features:list[str] = data['features']

	async def getMore(self) -> 'PartialGuild':
		return await self.instance.http.get_guild(self.id)

	async def getFull(self) -> 'Guild':
		for guild in self.instance.guilds:
			if guild.id == self.id:
				return guild
		raise Exception("Client is not in guild, please use the \"getMore\" function to get guild info instead. (will not include channels, emojis, roles, and members/membercount)")

	async def leave(self) -> None:
		return await self.instance.http.leave_guild(self.id)

	async def delete(self) -> None:
		return await self.instance.http.delete_guild(self.id)

	async def kick(self,id:str,reason:Union[str,None]=None) -> None:
		return await self.instance.http.kick(self.id,id,reason)

	async def getBans(self) -> List['Ban']:
		return await self.instance.http.get_bans(self.id)

	async def banMember(self,user_id:str,reason:Union[str,None]=None,delete_message_days:int=0) -> None:
		return await self.instance.http.ban(self.id,user_id,reason,delete_message_days)

	async def unbanMember(self,user_id:str) -> None:
		return await self.instance.http.unban(self.id,user_id)

	async def updateMember(self,id:str,nick:Union[str,None]=None,roles:Union[List[Union[Role,str]],None]=None)-> None:
		return await self.instance.http.update_guild_member(self.id,id,nick,roles)

	async def update(self,name:Union[str,None]=None,icon:Union[File,None]=None) -> 'PartialGuild':
		return await self.instance.http.update_guild(self.id,name,icon)

	async def nickSelf(self,nick:str=""):
		return await self.instance.http.update_client_nick(self.id,nick)

	async def getChannel(self,channel_id:str) -> 'Channel':
		return await self.instance.http.get_channel(channel_id)

	async def updateChannelPostion(self,channel_id:str,position:int) -> List['Channel']:
		return await self.instance.http.update_channel_position(self.id,channel_id,position)

	async def updateChannel(self,id:str,name:Union[str,None]=None,topic:Union[str,None]=None,nsfw:Union[bool,None]=None,bitrate:Union[int,None]=None,user_limit:Union[int,None]=None) -> 'Channel':
		return await self.instance.http.update_channel(self.id,id,name,topic,nsfw,bitrate,user_limit)

	async def createChannel(self,name:str,type:int=0,bitrate:Union[int,None]=None,user_limit:Union[int,None]=None,permission_overwrites:Union[List[PermissionOverwrite],None]=None,nsfw:bool=False) -> 'Channel':
		return await self.instance.http.create_channel(self.id,name,type,bitrate,user_limit,permission_overwrites,nsfw)

	async def deleteChannel(self,id:str) -> None:
		return await self.instance.http.delete_channel(id)

	async def createRole(self) -> Role:
		return await self.instance.http.create_role(self.id)

	async def setRolePostion(self,id:str,position:int) -> List[Role]:
		return await self.instance.http.set_role_position(self.id,id,position)

	async def editRole(self,guild_id:str,id:str,name:Union[str,None]=None,permissions:Union[Permissions,int,None]=None,color:Union[int,None]=None,hoist:Union[bool,None]=None,mentionable:Union[bool,None]=None) -> Role:
		return await self.instance.http.edit_role(guild_id,id,name,permissions,color,hoist,mentionable)
	
	async def deleteRole(self,id:str) -> None:
		return await self.instance.http.delete_role(self.id,id)

	async def getEmojis(self) -> List['Emoji']:
		return await self.instance.http.get_emojis(self.id)

	async def getEmoji(self,id:str) -> 'Emoji':
		return await self.instance.http.get_emoji(self.id,id)

	async def createEmoji(self,name:str,image:File) -> 'Emoji':
		return await self.instance.http.create_emoji(self.id,name,image)

	async def editEmoji(self,id:str,name:str) -> 'Emoji':
		return await self.instance.http.edit_emoji(self.id,id,name)

	async def deleteEmoji(self,id:str) -> None:
		return await self.instance.http.delete_emoji(self.id,id)
	
	async def createInvite(self) -> 'Invite':
		return await self.instance.http.create_channel_invite(self.id) #default channel id is the same as guild id

	async def getChannelInvites(self,id:str) -> List['Invite']:
		return await self.instance.http.get_channel_invites(id)

class Typing:
	def __init__(self,channel):
		self._isTyping = False
		self.channel = channel

	async def doTypingAction(self):
		while self._isTyping:
			await self.channel.startTyping()
			await asyncio.sleep(10)

	async def __aenter__(self):
		self._isTyping = True
		self.background_task = asyncio.create_task(self.doTypingAction())
		return self

	async def __aexit__(self, exc_type, exc_val, exc_tb):
		self._isTyping = False
		if self.background_task:
			self.background_task.cancel()
			with contextlib.suppress(asyncio.CancelledError):
				await self.background_task

class PartialChannel:
	def __init__(self,data,guild_id,instance):
		self.instance = instance
		self.guild_id:Union[str,None] = guild_id
		self.id:str = data['id']
		self.type:int = data['type']
		self.name:Union[str,None] = data.get('name') #text and gc
		self.typing:Typing = Typing(self)

	async def getFull(self) -> 'Channel':
		return await self.instance.http.get_channel(self.id)

	async def getOverwrites(self) -> List[PermissionOverwrite]:
		for guild in self.instance.guilds:
			if guild.id == self.guild_id:
				for channel in guild.channels:
					if channel.id == self.id:
						return channel.overwrites
		raise Exception("Client is not in a guild that contains this channel!")

	async def startTyping(self) -> None:
		return await self.instance.http.start_typing(self.id)

	async def bulkDelete(self,messages:List[Union[Message,str]]) -> None:
		return await self.instance.http.bulk_delete(self.id,messages)

	async def sendMessage(self,message:str="",embeds:List[Embed]=[],file:Union[File,None]=None,tts:bool=False) -> Message:
		return await self.instance.http.send_message(self.id,message,embeds,file,tts)

	async def getMessages(self,limit:int=50) -> List[Message]:
		return await self.instance.http.get_messages(self.id,limit)

	async def getPins(self) -> List[Message]:
		return await self.instance.http.get_channel_pins(self.id)

	async def pinMessage(self,message:Union[Message,str]) -> None:
		message = message.id if type(message) == Message else message
		return await self.instance.http.pin_message(self.id,message)

	async def unpinMessage(self,message:Union[Message,str]) -> None:
		message = message.id if type(message) == Message else message
		return await self.instance.http.unpin_message(self.id,message)

	async def getInvites(self) -> List['Invite']:
		return await self.instance.http.get_channel_invites(self.id)

	async def createInvite(self,max_age:Union[int,None]=None,max_uses:Union[int,None]=None,temporary:Union[bool,None]=None,unique:Union[bool,None]=None) -> 'Invite':
		return await self.instance.http.create_channel_invite(self.id,max_age,max_uses,temporary,unique)

	async def updatePosition(self,position:int) -> List['Channel']:
		return await self.instance.http.update_channel_position(self.guild_id,self.id,position)

	async def updateChannel(self,name:Union[str,None]=None,position:Union[int,None]=None,topic:Union[str,None]=None,nsfw:Union[bool,None]=None,bitrate:Union[int,None]=None,user_limit:Union[int,None]=None) -> 'Channel':
		if not name:
			name = self.name
		return await self.instance.http.update_channel(self.id,name,position,topic,nsfw,bitrate,user_limit)

	async def addOverwrite(self,overwrite:PermissionOverwrite) -> None:
		return await self.instance.http.add_channel_overwrite(self.id,overwrite)

	async def deleteOverwrite(self,target:str) -> None:
		return await self.instance.http.delete_channel_overwrite(self.id,target)

	async def delete(self) -> None:
		return await self.instance.http.delete_channel(self.id)

class Invite:
	def __init__(self,data,instance,has_metadata:bool=False):
		from .user import User
		self.instance = instance
		self.url:str = self.instance.base_url.replace("/api","").replace("/v6","") + data['code']
		self.code:str = data['code']
		self.guild:InvitePartialGuild = InvitePartialGuild(data['guild'],instance)
		self.channel:PartialChannel = PartialChannel(data['channel'],data['guild']['id'],instance)
		self.approximate_presence_count:Union[int,None] = data.get('approximate_presence_count')
		self.approximate_member_count:Union[int,None] = data.get('approximate_member_count')
		self.has_metadata:bool = has_metadata
		self.created_at:Union[str,None] = data.get('created_at')
		self.max_uses:Union[int,None] = data.get('max_uses')
		self.max_age:Union[int,None] = data.get('max_age')
		self.temporary:Union[bool,None] = data.get('temporary')
		self.uses:Union[int,None] = data.get('uses')
		self.inviter:User = User(data['inviter'],instance,self.guild.id)

	async def delete(self) -> None:
		return await self.instance.http.delete_invite(self.code)

class PartialGuild(InvitePartialGuild):
	def __init__(self,data,instance):
		super().__init__(data,instance)
		self.region = data['region']
		self.roles = [Role(role,self.id,instance) for role in data['roles']]
		self.emojis = [Emoji(emoji,instance) for emoji in data['emojis']]
		self.voice_states = data['voice_states']
		self.afk_timeout = data['afk_timeout']

	async def getMore(self) -> 'PartialGuild': #not needed as it will just return this object again
		raise Exception("PartialGuild.getMore() is not needed as it will just return this object again.")

class Channel(PartialChannel):
	def __init__(self,data,instance):
		from .user import User
		super().__init__(data,data['guild_id'],instance)
		self.instance = instance
		self.guild_id:Union[str,None] = data['guild_id'] #always exists
		self.last_message_id:Union[str,None] = data.get('last_message_id') #text-based
		self.topic:Union[str,None] = data.get('topic') #text
		self.nsfw:Union[bool,None] = bool(data.get('nsfw')) #text
		self.bitrate:Union[int,None] = data.get('bitrate') #voice
		self.user_limit:Union[int,None] = data.get('user_limit') #voice
		self.owner_id:Union[str,None] = data.get('owner_id') #gc
		self.application_id:Union[str,None] = data.get('application_id') #gc
		if self.guild_id:
			self.permission_overwrites:list[PermissionOverwrite] = [PermissionOverwrite(overwrite) for overwrite in data['permission_overwrites']]
			self.position:int = data['position']
		if data['type'] == 3:
			self.recipients:list[User] = [User(data,instance) for data in data['recipients']]
			self.icon:Icon = Icon(data,"icon",self.instance)
		if data['type'] == 1:
			self.recipients:list[User] = [User(data,instance) for data in data['recipients']]

	async def getFull(self) -> 'Channel': #not needed as it will just return this object again
		raise Exception("Channel.getFull() is not needed as it will just return this object again.")

	def toDict(self):
		if self.type == 0:
			return {"name":self.name,"permission_overwrites":self.permission_overwrites,"position":self.position,"topic":self.topic,"nsfw":self.nsfw,"last_message_id":self.last_message_id,"type":self.type,"id":self.id,"guild_id":self.guild_id}
		elif self.type == 2:
			return {"bitrate":self.bitrate,"name":self.name,"permission_overwrites":self.permission_overwrites,"position":self.position,"user_limit":self.user_limit,"type":self.type,"id":self.id,"guild_id":self.guild_id}

	async def updateChannel(self,name:Union[str,None]=None,topic:Union[str,None]="",nsfw:Union[bool,None]=None,bitrate:Union[int,None]=None,user_limit:Union[int,None]=None) -> 'Channel': #type:ignore | fills in info not available in a PartialChannel object
		if not name:
			name = self.name
		if self.type == 0:
			if topic == "":
				topic = self.topic
			if nsfw == None:
				nsfw = self.nsfw
		if self.type == 2:
			if not bitrate:
				bitrate = self.bitrate
			if not user_limit:
				user_limit = self.user_limit
		return await self.instance.http.update_channel(self.id,name,topic,nsfw,bitrate,user_limit)

class Emoji:
	def __init__(self,data,instance):
		from .user import User
		self.instance = instance
		self.id:str = data['id']
		self.name:str = data['name']
		self.guild_id:str = data.get('guild_id')
		self.roles:list = data['roles']
		self.user:User = User(data['user'],instance,self.guild_id)
		self.require_colons:bool = data['require_colons']
		self.managed:bool = data['managed']
		self.allNamesString:str = data['allNamesString']

	async def edit(self,name) -> 'Emoji':
		return await self.instance.http.edit_emoji(self.guild_id,self.id,name)

	async def delete(self) -> None:
		return await self.instance.http.delete_emoji(self.guild_id,self.id)

	def toDict(self):
		return {"id":self.id,"name":self.name,"roles":self.roles,"user":self.user.toDict(),"require_colons":self.require_colons,"managed":self.managed,"allNamesString":self.allNamesString}

class Guild(PartialGuild):
	def __init__(self,guild,instance):
		from .user import Member, Presence
		super().__init__(guild,instance)
		self.instance = instance
		self.channels:list[Channel] = [Channel(channel,instance) for channel in guild['channels']]
		self.members:list[Member] = [Member(member,self.id,instance,self.roles) for member in guild['members']]
		self.owner:Member = None #type:ignore
		for member in self.members:
			if member.id == self.owner_id:
				self.owner = member
		self.large:bool = guild['large']
		self.member_count:int = guild['member_count']
		self.presences:list[Presence] = [Presence(presence,self.id,instance) for presence in guild['presences']]
		self.unavailable:Union[bool,None] = guild.get('unavailable') #nonexistent on guild join mid-session

	async def getFull(self) -> 'Guild': #not needed as it will just return this object again
		raise Exception("Guild.getFull() is not needed as it will just return this object again.")

	async def getClientMember(self):
		for member in self.members:
			if member.id == self.instance.state.user.id:
				return member
		raise Exception("Client is not in this guild!")

	async def getMember(self,id):
		for member in self.members:
			if id == member.id:
				return member
		raise Exception("Could not find member in guild!")

	async def getRole(self,id) -> Role:
		for role in self.roles:
			if id == role.id:
				return role
		raise Exception("Could not find role in guild!")

	def _update_data(self,guild:PartialGuild):
		self.name = guild.name
		self.region = guild.region
		self.icon = guild.icon
		self.owner_id = guild.owner_id
		self.afk_timeout = guild.afk_timeout
		self.features = guild.features
		self.voice_states = guild.voice_states
		self.roles = guild.roles
		self.emojis = guild.emojis

	def toDict(self):
		return {"channels":[channel.toDict() for channel in self.channels], "emojis":[emoji.toDict() for emoji in self.emojis], "roles":[role.toDict() for role in self.roles], "members":[member.toDict() for member in self.members], "large":self.large, "member_count":self.member_count, "presences":[presence.toDict() for presence in self.presences],"unavailable":self.unavailable}

class Ban:
	def __init__(self,data,guild_id,instance):
		from .user import User
		self.instance = instance
		self.guild_id:str = guild_id
		self.reason:Union[str,None] = data['reason']
		self.user:User = User(data['user'],instance)

	async def revoke(self) -> None:
		return await self.instance.http.unban(self.guild_id,self.user.id)



from ..guild import Guild,Channel,Typing,Invite
from ..user import User,Member
from ..message import Message
from ..internal import Reply
from ..embed import Embed
from ..file import File

from typing import *
import re

class Context:
	def __init__(self,data,instance):
		from ..http import HTTP
		self.instance = instance
		self.message:Message = data
		self.channel:Channel = data.channel
		self.guild:Union[Guild,None] = data.guild
		self.guild_id:str = data.guild_id
		self.channel_id:str = data.channel_id
		self.author:User = data.author
		self.http:HTTP = instance.http
		self.typing:Typing = Typing(self.channel)

	async def reply(self, content:str="",embeds:List[Embed]=[],file:Union[File,None]=None) -> Message:
		filtered = re.sub(r'<@!?(\d+)>|@everyone|<@&(\d+)>', '[mention]', self.message.content)
		return await self.http.send_message(self.message.channel_id,f"> {filtered} \n<@{self.author.id}> {content}",embeds,file,_reply=Reply(self.message))

	async def sendMessage(self, content:str="",embeds:List[Embed]=[],file:Union[File,None]=None) -> Message:
		return await self.http.send_message(self.message.channel_id,content,embeds,file)

	async def startTyping(self) -> None:
		return await self.http.start_typing(self.channel.id)

	async def getPins(self):
		return await self.http.get_channel_pins(self.channel.id)

	async def createInvite(self) -> Invite:
		return await self.http.create_channel_invite(self.channel.id)

	async def getClientMember(self) -> Member:
		if self.guild: #people can use ctx.state.user if they need client user and the received message results to not be in a guild
			return await self.guild.getClientMember()
		raise Exception("Client is not in a guild!")



from ...guild import PartialGuild, Ban, Emoji
from ...role import Role, Permissions
from ...user import Member
from ...file import File

from typing import *

import base64

class hGuild:
	def __init__(self,instance):
		self.instance = instance
		self.s = instance.s

	async def get_guild(self,id:str) -> PartialGuild:
		from ... import HTTPStatus
		r = self.s.get(url=f"{self.instance.base_url}guilds/{id}")
		s = HTTPStatus(r)
		if s.success:
			return PartialGuild(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def leave_guild(self,guild_id:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(f"{self.instance.base_url}users/@me/guilds/{guild_id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def delete_guild(self,guild_id:str) -> None:
		from ... import HTTPStatus
		r = self.s.post(f"{self.instance.base_url}guilds/{guild_id}/delete") #clowning on the discord docs rn cause theyre WRONG (they said it was DELETE /guilds/{guild_id})
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def kick(self,guild_id:str,user_id:str,reason:Union[str,None]=None) -> None:
		from ... import HTTPStatus
		self.s.headers['X-Audit-Log-Reason'] = reason
		r = self.s.delete(url=f"{self.instance.base_url}guilds/{guild_id}/members/{user_id}")
		self.s.headers['X-Audit-Log-Reason'] = None
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def get_bans(self,guild_id:str) -> List[Ban]:
		from ... import HTTPStatus
		r = self.s.get(url=f"{self.instance.base_url}/guilds/{guild_id}/bans")
		s = HTTPStatus(r)
		if s.success:
			return [Ban(ban,guild_id,self.instance) for ban in r.json()]
		else:
			raise s.exception(s.reason)

	async def ban(self,guild_id:str,user_id:str,reason:Union[str,None]=None,delete_message_days:int=0) -> None:
		from ... import HTTPStatus
		self.s.headers['X-Audit-Log-Reason'] = reason
		r = self.s.put(url=f"{self.instance.base_url}guilds/{guild_id}/bans/{user_id}",json={'delete-message-days':int(delete_message_days),'reason':reason})
		self.s.headers['X-Audit-Log-Reason'] = None
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def unban(self,guild_id:str,user_id:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(url=f"{self.instance.base_url}guilds/{guild_id}/bans/{user_id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def update_guild_member(self,guild_id:str,user_id:str,nick:Union[str,None]=None,roles:Union[List[Union[Role,str]],None]=None) -> None: #mute,deaf,channel_id params are for voice, which doesn't exist on hummus yet
		from ... import HTTPStatus
		data = {}
		if roles:
			data['roles'] = [role.id if type(role) == Role else role for role in roles]
		if nick:
			data['nick'] = nick
		r = self.s.patch(url=f"{self.instance.base_url}guilds/{guild_id}/members/{user_id}",json=data)
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def create_guild(self,name:str,region:Union[str,None]=None,icon:Union[File,None]=None) -> PartialGuild:
		data = {}
		if icon:
			data["icon"] = f"data:image/png;base64,{base64.b64encode(await icon.get_file_data()).decode('utf-8')}"
		if region:
			data['region'] = region
		data['name'] = name
		from ... import HTTPStatus
		r = self.s.post(url=f"{self.instance.base_url}guilds",json=data)
		s = HTTPStatus(r)
		if s.success:
			return PartialGuild(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def update_client_nick(self,guild_id:str,nick:str="") -> Member:
		from ... import HTTPStatus
		if not nick: #user attempts to pass in None to reset nick
			nick = ""
		r = self.s.patch(url=f"{self.instance.base_url}guilds/{guild_id}/members/@me/nick",json={"nick":nick})
		s = HTTPStatus(r)
		if s.success:
			return Member(r.json(),guild_id,self.instance)
		else:
			raise s.exception(s.reason)

	async def delete_role(self,guild_id:str,id:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(url=f"{self.instance.base_url}guilds/{guild_id}/roles/{id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def create_role(self,guild_id:str) -> Role:
		from ... import HTTPStatus
		r = self.s.post(url=f"{self.instance.base_url}guilds/{guild_id}/roles") #when the json doesnt fucking work so people have to rely on Role.edit() lol! (fuck you ziad)
		s = HTTPStatus(r)
		if s.success:
			return Role(r.json(),guild_id,self.instance)
		else:
			raise s.exception(s.reason)

	async def modify_role(self,guild_id:str,id:str,name:Union[str,None]=None,permissions:Union[Permissions,int,None]=None,color:Union[int,None]=None,hoist:Union[bool,None]=None,mentionable:Union[bool,None]=None) -> Role:
		data = {}
		if name:
			data['name'] = name
		if permissions:
			if type(permissions) == Permissions:
				permissions.update()
				permissions = permissions.value
			data['permissions'] = permissions
		if color:
			data['color'] = int(color)
		if hoist != None:
			data['hoist'] = hoist
		if mentionable != None:
			data['mentionable'] = mentionable
		from ... import HTTPStatus
		r = self.s.patch(url=f"{self.instance.base_url}guilds/{guild_id}/roles/{id}",json=data)
		s = HTTPStatus(r)
		if s.success:
			return Role(r.json(),guild_id,self.instance)
		else:
			raise s.exception(s.reason)

	async def modify_role_position(self,guild_id:str,id:str,position:int) -> List[Role]:
		from ... import HTTPStatus
		r = self.s.patch(url=f"{self.instance.base_url}guilds/{guild_id}/roles",json=[{"id":id,"position":int(position)}])
		s = HTTPStatus(r)
		if s.success:
			return [Role(role,guild_id,self.instance) for role in r.json()]
		else:
			raise s.exception(s.reason)

	async def modify_guild(self,guild_id:str,name:Union[str,None]=None,icon:Union[File,None]=None) -> PartialGuild:
		data = {}
		if name:
			data['name'] = name
		if icon:
			if not icon.empty:
				data["icon"] = f"data:image/png;base64,{base64.b64encode(await icon.get_file_data()).decode('utf-8')}"
		else:
			e = self.s.get(f"{self.instance.base_url}guilds/{guild_id}").json()
			icondata = self.s.get(f"{self.instance.cdn}icons/{e['id']}/{e['icon']}.png").content
			data["icon"] = f"data:image/png;base64,{base64.b64encode(icondata).decode('utf-8')}"
		from ... import HTTPStatus
		r = self.s.patch(url=f"{self.instance.base_url}guilds/{guild_id}",json=data)
		s = HTTPStatus(r)
		if s.success:
			return PartialGuild(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def get_emojis(self,guild_id:str) -> List[Emoji]:
		from ... import HTTPStatus
		r = self.s.get(f"{self.instance.base_url}guilds/{guild_id}/emojis")
		s = HTTPStatus(r)
		if s.success:
			return [Emoji(emoji,self.instance) for emoji in r.json()]
		else:
			raise s.exception(s.reason)

	async def get_emoji(self,guild_id:str,id:str) -> Emoji:
		from ... import HTTPStatus
		r = self.s.get(f"{self.instance.base_url}guilds/{guild_id}/emojis/{id}")
		s = HTTPStatus(r)
		if s.success:
			return Emoji(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def create_emoji(self,guild_id:str,name:str,image:File) -> Emoji:
		from ... import HTTPStatus
		r = self.s.post(f"{self.instance.base_url}guilds/{guild_id}/emojis",json={"name":name,"image":f"data:image/png;base64,{base64.b64encode(await image.get_file_data()).decode('utf-8')}"})
		s = HTTPStatus(r)
		if s.success:
			return Emoji(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def edit_emoji(self,guild_id:str,id:str,name:str) -> Emoji:
		from ... import HTTPStatus
		r = self.s.patch(f"{self.instance.base_url}guilds/{guild_id}/emojis/{id}",json={"name":name})
		s = HTTPStatus(r)
		if s.success:
			return Emoji(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def delete_emoji(self,guild_id:str,id:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(url=f"{self.instance.base_url}guilds/{guild_id}/emojis/{id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)



from ...user import User, Profile

class hUser:
	def __init__(self,instance):
		self.instance = instance
		self.s = instance.s

	async def get_user(self,id:str) -> User:
		from ... import HTTPStatus
		r = self.s.get(url=f"{self.instance.base_url}users/{id}/")
		s = HTTPStatus(r)
		if s.success:
			return User(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def get_user_profile(self,id:str) -> Profile:
		from ... import HTTPStatus
		r = self.s.get(url=f"{self.instance.base_url}users/{id}/profile")
		s = HTTPStatus(r)
		if s.success:
			return Profile(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

"""
	async def create_dm(self,id):
		r = self.s.post(url=f"{self.instance.base_url}users/@me/channels" ,json={"recipient_id":id})
		print(r)
		print(r.json())
""" #returns 400, does not work at all, cannot figure out how to make it work


from typing import *
import tempfile
import filetype
import io

class File:
	def __init__(self,file:Union[str,bytes,io.BufferedReader,io.BytesIO],file_name:str="file",empty:bool=False):
		self.empty = empty
		if empty:
			return
		read = None
		filename = None
		thing = None #me smol brain hav no idea what to name

		if type(file) == io.BufferedReader:
			file = file.read()
			with tempfile.NamedTemporaryFile(delete=True) as temp_file:
				temp_file.write(file)
				kind = filetype.guess(temp_file.name)
			filename = file_name
			read = file
			if not kind:
				thing = (file, read, 'unsupported')
			else:
				thing = (file, read, kind.mime)
				filename += f".{kind.extension}"

		elif type(file) == bytes:
			with tempfile.NamedTemporaryFile(delete=True) as temp_file:
				temp_file.write(file)
				kind = filetype.guess(temp_file.name)
			filename = file_name
			read = file
			if not kind:
				thing = (file, read, 'unsupported')
			else:
				thing = (file, read, kind.mime)
				filename += f".{kind.extension}"

		elif type(file) == str:
			openedfile = open(file,"rb")
			read = openedfile.read()
			kind = filetype.guess(file)
			filename = file.split("/")[-1]
			thing = (file, read, 'unsupported') if not kind else (file, read, kind.mime)

		elif type(file) == io.BytesIO:
			kind = filetype.guess(file.getvalue())
			filename = file_name
			read = file.getvalue()
			if not kind:
				thing = (file, read, 'unsupported')
			else:
				thing = (file, read, kind.mime)
				filename += f".{kind.extension}"

		else:
			from .utils import Exceptions
			raise Exceptions.FileError(f"Invalid file type: {type(file)}. Please pass a file object, bytes, BytesIO or string.")

		self.fields = {"file":thing,"payload_json":None}
		self.file_json = {"filename":filename,"Content-Type":thing[2]}

	async def get_file_data(self) -> bytes:
		if self.empty:
			return #type:ignore
		return self.fields['file'][1]

class Attachment:
	def __init__(self,data,instance):
		self.instance = instance
		self.id:str = data['id']
		self.filename:str = data['filename']
		self.size:int = data['size']
		self.url:str = data['url']
		self.proxy_url:str = data['proxy_url']
		self.height:int = data.get('height')
		self.width:int = data.get('width')

class Icon: #guild icon, user avatar, banner, etc
	def __init__(self,data,type,instance,is_user=False):
		from .utils import Enums #haha i love circular import errors!
		self.instance = instance
		self.object_id: str = data['id']
		self.hash: str = data.get(type)
		self.attachment_type: str = "gif" if self.hash and self.hash.startswith("a_") else "png"
		self.url: Union[str,None] = f"{instance.cdn+type}s/{self.object_id}/{self.hash}.{self.attachment_type}"
		if not self.hash:
			self.url = None
			if is_user:
				self.url = Enums.DefaultAvatars.all[int(data['discriminator'])%5]



from ...guild import Invite, Channel, PermissionOverwrite
from ...message import Message
from ...file import File

from typing import *
import base64

class hChannel:
	def __init__(self,instance):
		self.instance = instance
		self.s = instance.s

	async def create_channel(self,guild_id:str,name:str,type:int=0,bitrate:Union[int,None]=None,user_limit:Union[int,None]=None,permission_overwrites:Union[List[PermissionOverwrite],None]=None,nsfw:bool=False) -> Channel: #parent_id excluded because creating categories doesn't exist
		from ... import HTTPStatus
		data = {"name":name,"type":type,"nsfw":nsfw}
		if bitrate:
			data['bitrate'] = bitrate
		if user_limit:
			data['user_limit'] = user_limit
		if permission_overwrites:
			data['permission_overwrites'] = [{"id":overwrite.id,"type":overwrite.type,"allow":overwrite.allow.value,"deny":overwrite.deny.value} for overwrite in permission_overwrites]
		r = self.s.post(f"{self.instance.base_url}guilds/{guild_id}/channels",json=data)
		s = HTTPStatus(r)
		if s.success:
			return Channel(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def get_channel(self,id:str) -> Channel:
		from ... import HTTPStatus
		r = self.s.get(f"{self.instance.base_url}channels/{id}")
		s = HTTPStatus(r)
		if s.success:
			return Channel(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def get_messages(self,channel_id:str,limit:int=50) -> List[Message]:
		from ... import HTTPStatus
		r = self.s.get(url=f"{self.instance.base_url}channels/{channel_id}/messages/",json={"limit":limit})
		s = HTTPStatus(r)
		if s.success:
			return [Message(message,self.instance) for message in r.json()]
		else:
			raise s.exception(s.reason)

	async def delete_channel(self,id:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(f"{self.instance.base_url}channels/{id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def get_channel_pins(self,id:str) -> List[Message]:
		from ... import HTTPStatus, Message
		r = self.s.get(f"{self.instance.base_url}channels/{id}/pins")
		s = HTTPStatus(r)
		if s.success:
			return [Message(message,self.instance) for message in r.json()]
		else:
			raise s.exception(s.reason)

	async def get_channel_invites(self,id:str) -> List[Invite]:
		from ... import HTTPStatus
		r = self.s.get(f"{self.instance.base_url}channels/{id}/invites")
		s = HTTPStatus(r)
		if s.success:
			return [Invite(inv,self.instance,True) for inv in r.json()]
		else:
			raise s.exception(s.reason)

	async def create_channel_invite(self,id:str,max_age:Union[int,None]=None,max_uses:Union[int,None]=None,temporary:bool=False,unique:bool=False) -> Invite:
		from ... import HTTPStatus
		data = {}
		if max_age:
			data['max_age'] = max_age
		if max_uses:
			data['max_uses'] = max_uses
		if temporary:
			data['temporary'] = temporary
		if unique:
			data['unique'] = unique
		r = self.s.post(f"{self.instance.base_url}channels/{id}/invites",json=data)
		s = HTTPStatus(r)
		if s.success:
			return Invite(r.json(),self.instance,True)
		else:
			raise s.exception(s.reason)

	async def get_invite(self,id:str,with_counts:bool=False) -> Invite:
		from ... import HTTPStatus
		r = self.s.get(f"{self.instance.base_url}invites/{id}",json={"with_counts":with_counts})
		s = HTTPStatus(r)
		if s.success:
			return Invite(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def delete_invite(self,id:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(f"{self.instance.base_url}invites/{id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def update_channel_position(self,guild_id:str,id:str,position:int) -> List[Channel]:
		from ... import HTTPStatus
		r = self.s.patch(f"{self.instance.base_url}guilds/{guild_id}/channels/",json=[{"id":id,"position":int(position)}])
		s = HTTPStatus(r)
		if s.success:
			return [Channel(channel,self.instance) for channel in r.json()]
		else:
			raise s.exception(s.reason)

	async def update_channel(self,id:str,name:Union[str,None]=None,topic:Union[str,None]=None,nsfw:Union[bool,None]=None,bitrate:Union[int,None]=None,user_limit:Union[int,None]=None) -> Channel:
		data = {}
		if name:
			data['name'] = name
		if topic:
			data['topic'] = topic
		if nsfw:
			data['nsfw'] = nsfw
		if bitrate:
			data['bitrate'] = bitrate
		if user_limit:
			data['user_limit'] = user_limit
		from ... import HTTPStatus
		r = self.s.patch(url=f"{self.instance.base_url}/channels/{id}",json=data)
		s = HTTPStatus(r)
		if s.success:
			return Channel(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def add_channel_overwrite(self,channel_id:str,overwrite:PermissionOverwrite) -> None:
		from ... import HTTPStatus
		if type(overwrite) != PermissionOverwrite:
			raise Exception("Please provide a Permissions Overwrite (PermissionOverwrite) object.")
		overwrite.allow.update()
		overwrite.deny.update()
		data = {"id":overwrite.id,"type":overwrite.type,"allow":overwrite.allow.value,"deny":overwrite.deny.value}
		r = self.s.put(url=f"{self.instance.base_url}channels/{channel_id}/permissions/{overwrite.id}",json=data)
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def delete_channel_overwrite(self,channel_id:str,target:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(f"{self.instance.base_url}channels/{channel_id}/permissions/{target}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)

	async def update_gc(self,id:str,name:Union[str,None]=None,icon:Union[File,None]=None) -> Channel:
		from ... import HTTPStatus
		data = {}
		if icon:
			data['icon'] = f"data:image/png;base64,{base64.b64encode(await icon.get_file_data()).decode('utf-8')}"
		if name:
			data['name'] = name
		r = self.s.patch(f"{self.instance.base_url}channels/{id}",json=data)
		s = HTTPStatus(r)
		if s.success:
			return Channel(r.json(),self.instance)
		else:
			raise s.exception(s.reason)

	async def leave_gc(self,id:str) -> None:
		from ... import HTTPStatus
		r = self.s.delete(f"{self.instance.base_url}channels/{id}")
		s = HTTPStatus(r)
		if not s.success:
			raise s.exception(s.reason)



#from .internal import *
from .message import *
from .embed import *
from .guild import *
from .ready import *
from .file import *
from .http import *
from .main import *
from .role import *
from .user import *
from . import utils

__version__ = "1.1.1"

#what the hell is "imports is un-sorted/formatted"!?!?!?!?



from .utils import Exceptions, CustomStatus
from .internal import http

import json

class HTTPStatus:
	def __init__(self,response):
		self.success = False
		self.response = response
		if 199 < response.status_code < 300: #i think some 300-range status codes can be considered "successful", but I don't think Hummus ever returns them
			self.success = True
			return
		if response.status_code == 400:
			self.reason = "Client formatted the request incorrectly!"
			self.exception = Exceptions.BadRequest
			return
		if response.status_code == 401:
			self.reason = "Client is not allowed to perform this action! (likely because invalid token)"
			self.exception = Exceptions.NotAllowed
			return
		if response.status_code == 403:
			self.reason = "Client does not have permissions to perform this action!"
			self.exception = Exceptions.InvalidPermissions
			return
		if response.status_code == 404:
			self.reason = "Client could not find the requested resource!"
			self.exception = Exceptions.NotFound
			return
		if response.status_code > 499:
			self.reason = "Server had an internal error!"
			self.exception = Exceptions.InternalServerError
			return
		try:
			if "unknown" in response.json()['message'].lower():
				self.reason = "Client could not find the requested resource!"
				self.exception = Exceptions.NotFound
				return
		except:
			pass
		self.reason = "An unknown error occurred!"
		self.exception = Exceptions.UnknownError

class HTTP(http.hChannel,http.hGuild,http.hMessage,http.hSelf,http.hUser):
	def __init__(self,instance):
		super().__init__(instance)

	#no need to create a whole file for one function
	async def update_status(self,status:CustomStatus) -> None:
		self.instance.connection.send(json.dumps({"op":3,"d":status._toJson()}))